@@ -1,6 +1,6 @@
 /* QuesoGLC
  * A free implementation of the OpenGL Character Renderer (GLC)
- * Copyright (c) 2002, 2004-2007, Bertrand Coconnier
+ * Copyright (c) 2002, 2004-2008, Bertrand Coconnier
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -16,7 +16,7 @@
  *  License along with this library; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
-/* $Id: font.c,v 1.77 2007/10/21 16:22:29 bcoconni Exp $ */
+/* $Id: font.c 710 2008-01-21 20:14:17Z bcoconni $ */
 
 /** \file
  *  defines the so-called "Font commands" described in chapter 3.7 of the GLC
@@ -115,9 +115,8 @@ void __glcAppendFont(__GLCcontext* inContext, __GLCfont* inFont)
     return;
   }
 
-#ifndef FT_CACHE_H
-  if (!__glcFaceDescOpen(inFont->faceDesc, inContext)) {
-    __glcRaiseError(GLC_RESOURCE_ERROR);
+#ifndef GLC_FT_CACHE
+  if (!__glcFontOpen(inFont, inContext)) {
     __glcFree(node);
     return;
   }
@@ -185,8 +184,8 @@ static void __glcDeleteFont(__GLCfont* font, __GLCcontext* inContext)
   /* If the font has been found, remove it from the list */
   if (node) {
     FT_List_Remove(&inContext->currentFontList, node);
-#ifndef FT_CACHE_H
-    __glcFaceDescClose(font->faceDesc);
+#ifndef GLC_FT_CACHE
+    __glcFontClose(font);
 #endif
     __glcFree(node);
   }
@@ -198,10 +197,12 @@ static void __glcDeleteFont(__GLCfont* font, __GLCcontext* inContext)
 /** \ingroup font
  *  This command deletes the font identified by \e inFont. If \e inFont is an
  *  element in the list \b GLC_CURRENT_FONT_LIST, the command removes that
- *  element from the list.
+ *  element from the list. All information about the font is lost, and the
+ *  indice \e inFont become unused.
  *
  *  The command raises \b GLC_PARAMETER_ERROR if \e inFont is not an element of
- *  the list \b GLC_FONT_LIST.
+ *  the list \b GLC_FONT_LIST and if \e inFont has not been generated by
+ *  glcGenFontID().
  *  \param inFont The ID of the font to delete
  *  \sa glcGetListi() with argument \b GLC_FONT_LIST
  *  \sa glcGeti() with argument \b GLC_FONT_COUNT
@@ -218,36 +219,57 @@ void APIENTRY glcDeleteFont(GLint inFont)
 
   GLC_INIT_THREAD();
 
-  /* Verify that the thread has a current context and that the font identified
-   * by 'inFont' exists.
-   */
-  font = __glcVerifyFontParameters(inFont);
-  if (!font)
-    return;
-
   ctx = GLC_GET_CURRENT_CONTEXT();
+  if (!ctx) {
+    __glcRaiseError(GLC_STATE_ERROR);
+    return;
+  }
+
+  /* Search the font to be deleted */
+  for (node = ctx->fontList.head; node; node = node->next) {
+    font = (__GLCfont*)node->data;
+    if (font->id == inFont) {
+      /* remove the font from the GLC_FONT_LIST then destroy it */
+      FT_List_Remove(&ctx->fontList, node);
+      break;
+    }
+  }
+
+  if (!node) {
+    for (node = ctx->genFontList.head; node; node = node->next) {
+      font = (__GLCfont*)node->data;
+      if (font->id == inFont) {
+	/* Remove the font from the empty font list generated by
+	 * glcGenFontID().
+	 */
+	FT_List_Remove(&ctx->genFontList, node);
+	break;
+      }
+    }
+
+    if (!node) {
+      __glcRaiseError(GLC_PARAMETER_ERROR);
+      return;
+    }
+  }
 
-  /* remove the font from the GLC_FONT_LIST then destroy it */
-  node = FT_List_Find(&ctx->fontList, font);
-  assert(node);
-  FT_List_Remove(&ctx->fontList, node);
   __glcFree(node);
   __glcDeleteFont(font, ctx);
 }
 
 
 
-#ifndef FT_CACHE_H
+#ifndef GLC_FT_CACHE
 /* Function called by FT_List_Finalize
  * Close the face of a font when GLC_CURRENT_FONT_LIST is deleted
  */
-static void __glcCloseFace(FT_Memory inMemory, void* data, void* user)
+static void __glcCloseFace(FT_Memory GLC_UNUSED_ARG(inMemory), void* data,
+			   void* GLC_UNUSED_ARG(user))
 {
   __GLCfont* font = (__GLCfont*)data;
 
   assert(font);
-  assert(font->faceDesc);
-  __glcFaceDescClose(font->faceDesc);
+  __glcFontClose(font);
 }
 #endif
 
@@ -300,11 +322,9 @@ void APIENTRY glcFont(GLint inFont)
       FT_List_Remove(&ctx->currentFontList, node);
     }
     else {
-#ifndef FT_CACHE_H
-      if (!__glcFaceDescOpen(font->faceDesc, ctx)) {
-	__glcRaiseError(GLC_RESOURCE_ERROR);
+#ifndef GLC_FT_CACHE
+      if (!__glcFontOpen(font, ctx))
 	return;
-      }
 #endif
 
       /* Append the font identified by inFont to GLC_CURRENT_FONT_LIST */
@@ -313,11 +333,11 @@ void APIENTRY glcFont(GLint inFont)
         /* We keep the first node of the current font list in order not to need
          * to create a new one to store the font identified by 'inFont'
          */
-#ifndef FT_CACHE_H
+#ifndef GLC_FT_CACHE
         __GLCfont* dummyFont = (__GLCfont*)node->data;
 
         /* Close the face of the font stored in the first node */
-	__glcFaceDescClose(dummyFont->faceDesc);
+	__glcFontClose(dummyFont);
 #endif
         /* Remove the first node of the list to prevent it to be deleted by
          * FT_List_Finalize().
@@ -328,8 +348,8 @@ void APIENTRY glcFont(GLint inFont)
         /* The list is empty, create a new node */
         node = (FT_ListNode)__glcMalloc(sizeof(FT_ListNodeRec));
         if (!node) {
-#ifndef FT_CACHE_H
-	  __glcFaceDescClose(font->faceDesc);
+#ifndef GLC_FT_CACHE
+	  __glcFontClose(font);
 #endif
           __glcRaiseError(GLC_RESOURCE_ERROR);
           return;
@@ -337,7 +357,7 @@ void APIENTRY glcFont(GLint inFont)
       }
     }
 
-#ifndef FT_CACHE_H
+#ifndef GLC_FT_CACHE
     /* Close the remaining fonts in GLC_CURRENT_FONT_LIST and empty the list */
     FT_List_Finalize(&ctx->currentFontList, __glcCloseFace,
 		     &__glcCommonArea.memoryManager, NULL);
@@ -352,7 +372,7 @@ void APIENTRY glcFont(GLint inFont)
   }
   else {
     /* Empties the list GLC_CURRENT_FONT_LIST */
-#ifndef FT_CACHE_H
+#ifndef GLC_FT_CACHE
     FT_List_Finalize(&ctx->currentFontList, __glcCloseFace,
                      &__glcCommonArea.memoryManager, NULL);
 #else
@@ -364,76 +384,6 @@ void APIENTRY glcFont(GLint inFont)
 
 
 
-/* This internal function tries to open the face file which name is identified
- * by 'inFace'. If it succeeds, it closes the previous face and stores the new
- * face attributes in the __GLCfont object "inFont". Otherwise, it leaves the
- * font unchanged. GL_TRUE or GL_FALSE are returned to indicate if the function
- * succeeded or not.
- */
-GLboolean __glcFontFace(__GLCfont* inFont, const GLCchar8* inFace,
-			__GLCcontext *inContext)
-{
-  __GLCfaceDescriptor *faceDesc = NULL;
-  __GLCmaster *master = NULL;
-  __GLCcharMap* newCharMap = NULL;
-
-  GLC_INIT_THREAD();
-
-  /* TODO : Verify if the font has already the required face activated */
-
-  master = __glcMasterCreate(inFont->parentMasterID, inContext);
-  if (!master) {
-    __glcRaiseError(GLC_RESOURCE_ERROR);
-    return GL_FALSE;
-  }
-
-  /* Get the face descriptor of the face identified by the string inFace */
-  faceDesc = __glcFaceDescCreate(master, inFace, inContext);
-  if (!faceDesc) {
-    __glcRaiseError(GLC_RESOURCE_ERROR);
-    __glcMasterDestroy(master);
-    return GL_FALSE;
-  }
-
-  newCharMap = __glcCharMapCreate(master);
-  if (!newCharMap) {
-    __glcRaiseError(GLC_RESOURCE_ERROR);
-    __glcMasterDestroy(master);
-    return GL_FALSE;
-  }
-
-  __glcMasterDestroy(master);
-
-#ifndef FT_CACHE_H
-  /* If the font belongs to GLC_CURRENT_FONT_LIST then open the font file */
-  if (FT_List_Find(&inContext->currentFontList, font)) {
-
-    /* Open the new face */
-    if (!__glcFaceDescOpen(faceDesc, inContext)) {
-      __glcRaiseError(GLC_RESOURCE_ERROR);
-      __glcCharMapDestroy(newCharMap);
-      return GL_FALSE;
-    }
-
-    /* Close the current face */
-    __glcFaceDescClose(font->faceDesc);
-  }
-#endif
-
-  /* Destroy the current charmap */
-  if (inFont->charMap)
-    __glcCharMapDestroy(inFont->charMap);
-
-  inFont->charMap = newCharMap;
-
-  __glcFaceDescDestroy(inFont->faceDesc, inContext);
-  inFont->faceDesc = faceDesc;
-
-  return GL_TRUE;
-}
-
-
-
 /** \ingroup font
  *  This command attempts to set the current face of the font identified by
  *  \e inFont to the face identified by the string \e inFace. Examples for
@@ -479,10 +429,8 @@ GLboolean APIENTRY glcFontFace(GLint inFont, const GLCchar* inFace)
   }
 
   UinFace = __glcConvertToUtf8(inFace, ctx->stringState.stringType);
-  if (!UinFace) {
-    __glcRaiseError(GLC_RESOURCE_ERROR);
+  if (!UinFace)
     return GL_FALSE;
-  }
 
   if (inFont) {
     GLboolean result = GL_FALSE;
@@ -521,13 +469,8 @@ GLboolean APIENTRY glcFontFace(GLint inFont, const GLCchar* inFace)
       assert(master);
 
       /* Get the face descriptor of the face identified by the string inFace */
-      faceDesc = __glcFaceDescCreate(master, UinFace, ctx);
+      faceDesc = __glcFaceDescCreate(master, UinFace, ctx, 0);
       __glcMasterDestroy(master);
-      if (!faceDesc) {
-	__glcRaiseError(GLC_RESOURCE_ERROR);
-	return GL_FALSE;
-      }
-
       if (!faceDesc) {
 	/* No face identified by UinFace has been found in the font */
 	__glcFree(UinFace);
@@ -597,15 +540,12 @@ void APIENTRY glcFontMap(GLint inFont, GLint inCode, const GLCchar* inCharName)
      * The result is stored into 'buffer'.
      */
     buffer = __glcConvertToUtf8(inCharName, ctx->stringState.stringType);
-    if (!buffer) {
-      __glcRaiseError(GLC_RESOURCE_ERROR);
+    if (!buffer)
       return;
-    }
 
     /* Retrieve the Unicode code from its name */
     error = __glcCodeFromName(buffer);
     if (error < 0) {
-      __glcRaiseError(GLC_PARAMETER_ERROR);
       __glcFree(buffer);
       return;
     }
@@ -614,7 +554,6 @@ void APIENTRY glcFontMap(GLint inFont, GLint inCode, const GLCchar* inCharName)
     /* Get the glyph that corresponds to the mapped code */
     glyph = __glcFaceDescGetGlyph(font->faceDesc, mappedCode, ctx);
     if (!glyph) {
-      __glcRaiseError(GLC_PARAMETER_ERROR);
       __glcFree(buffer);
       return;
     }
@@ -628,7 +567,8 @@ void APIENTRY glcFontMap(GLint inFont, GLint inCode, const GLCchar* inCharName)
 
 /** \ingroup font
  *  This command returns a font ID that is not an element of the list
- *  \b GLC_FONT_LIST.
+ *  \b GLC_FONT_LIST. This command has also the effect of creating an empty
+ *  font for the returned ID so that this ID become used.
  *  \return A new font ID
  *  \sa glcDeleteFont()
  *  \sa glcIsFont()
@@ -638,6 +578,9 @@ void APIENTRY glcFontMap(GLint inFont, GLint inCode, const GLCchar* inCharName)
 GLint APIENTRY glcGenFontID(void)
 {
   __GLCcontext *ctx = NULL;
+  FT_ListNode node = NULL;
+  __GLCfont* font = NULL;
+  GLint id = 1;
 
   GLC_INIT_THREAD();
 
@@ -648,30 +591,49 @@ GLint APIENTRY glcGenFontID(void)
     return 0;
   }
 
-  /* The context state maintains the lower ID that is greater than all the
-   * the IDs of the fonts in GLC_FONT_LIST. So we return this ID and increments
-   * 'lastFontID' in order to take into account the new ID returned to the
-   * user.
-   * Notice however that :
-   * - Yes, this is a lazy policy
-   * - Even if the user never creates a font with this ID, QuesoGLC will
-   *   assume that this ID is lost until the context is deleted. This may lead
-   *   to unused font IDs.
-   * - If the user calls glcGenFontID() twice in a row he will get 2 different
-   *   IDs even though he/she has not used the first one. This may not be the
-   *   expected behaviour.
-   * - We do not take into account the overfow of the lastFontID parameter
-   *   and an evil user can make QuesoGLC returns the ID of a font that is
-   *   already a member of GLC_FONT_LIST
-   * But that would be the point in checking for such events ? Even if
-   * glcGenFontID() returns silly IDs, *every* GLC function that requires a
-   * font ID as a parameter checks its validity before using it. So yes, this
-   * function is really stupid but since it does not harm, let's keep it this
-   * way. In most cases, it will behave in a sensible way.
-   * Moreover, this behaviour is compliant with GLC specs and as a last resort
-   * the client can manage the font IDs itself.
-   */
-  return ctx->lastFontID++;
+  for (node = ctx->fontList.head; node; node = node->next) {
+    font = (__GLCfont*)node->data;
+
+    if (font->id == id) {
+      id++;
+      if (id > 0x7fffffff) {
+	__glcRaiseError(GLC_RESOURCE_ERROR);
+	return 0;
+      }
+      node = ctx->fontList.head;
+    }
+  }
+
+  for (node = ctx->genFontList.head; node; node = node->next) {
+    font = (__GLCfont*)node->data;
+
+    if (font->id == id) {
+      id++;
+      if (id > 0x7fffffff) {
+	__glcRaiseError(GLC_RESOURCE_ERROR);
+	return 0;
+      }
+      node = ctx->genFontList.head;
+    }
+  }
+
+  node = (FT_ListNode)__glcMalloc(sizeof(FT_ListNodeRec));
+  if (!node) {
+    __glcRaiseError(GLC_RESOURCE_ERROR);
+    return 0;
+  }
+
+  /* Create an empty font */
+  font = __glcFontCreate(id, NULL, ctx, 0);
+  if (!font) {
+    __glcFree(node);
+    return 0;
+  }
+
+  node->data = font;
+  FT_List_Add(&ctx->genFontList, node);
+
+  return id;
 }
 
 
@@ -697,10 +659,8 @@ const GLCchar* APIENTRY glcGetFontFace(GLint inFont)
     /* Convert the string name of the face into the current string type */
     buffer = __glcConvertFromUtf8ToBuffer(ctx, __glcFaceDescGetStyleName(font->faceDesc),
 					  ctx->stringState.stringType);
-    if (!buffer) {
-      __glcRaiseError(GLC_RESOURCE_ERROR);
+    if (!buffer)
       return GLC_NONE;
-    }
 
     /* returns the name */
     return buffer;
@@ -974,6 +934,15 @@ GLboolean APIENTRY glcIsFont(GLint inFont)
       return GL_TRUE;
   }
 
+  /* If the ID corresponds to an empty font reserved by glcGenFontID() then
+   * return GL_TRUE in order to identify the ID as reserved.
+   */
+  for (node = ctx->genFontList.head; node; node = node->next) {
+    __GLCfont* font = (__GLCfont*)node->data;
+    if (font->id == inFont)
+      return GL_TRUE;
+  }
+
   return GL_FALSE;
 }
 
@@ -983,39 +952,52 @@ GLboolean APIENTRY glcIsFont(GLint inFont)
  * creates a new font based on the pattern 'inPattern'. The resulting font is
  * added to the list GLC_FONT_LIST.
  */
-__GLCfont* __glcNewFontFromMaster(__GLCfont* inFont, GLint inFontID,
-				  __GLCmaster* inMaster,
-				  __GLCcontext *inContext)
+__GLCfont* __glcNewFontFromMaster(GLint inFontID, __GLCmaster* inMaster,
+				  __GLCcontext *inContext, GLint inCode)
 {
   FT_ListNode node = NULL;
   __GLCfont* font = NULL;
 
-  GLC_INIT_THREAD();
-
-  /* Create a new entry for GLC_FONT_LIST */
-  node = (FT_ListNode)__glcMalloc(sizeof(FT_ListNodeRec));
-  if (!node) {
-    __glcRaiseError(GLC_RESOURCE_ERROR);
-    return NULL;
+  /* Look for the font which ID is inFont in the GLC_FONT_LIST */
+  for (node = inContext->fontList.head; node; node = node->next) {
+    font = (__GLCfont*)node->data;
+    if (font->id == inFontID) {
+      FT_List_Remove(&inContext->fontList, node);
+      break;
+    }
   }
 
-  if (inFont)
-    /* Delete the font */
-    __glcDeleteFont(inFont, inContext);
+  if (!node) {
+    /* Look if the ID is in the empty fonts generated by glcGenFontID() */
+    for (node = inContext->genFontList.head; node; node = node->next) {
+      font = (__GLCfont*)node->data;
+      if (font->id == inFontID) {
+	FT_List_Remove(&inContext->genFontList, node);
+	break;
+      }
+    }
+  }
+
+  if (node)
+    __glcDeleteFont(font, inContext);
+  else {
+    /* Create a new entry for GLC_FONT_LIST */
+    node = (FT_ListNode)__glcMalloc(sizeof(FT_ListNodeRec));
+    if (!node) {
+      __glcRaiseError(GLC_RESOURCE_ERROR);
+      return NULL;
+    }
+  }
 
   /* Create a new font and add it to the list GLC_FONT_LIST */
-  font = __glcFontCreate(inFontID, inMaster, inContext);
+  font = __glcFontCreate(inFontID, inMaster, inContext, inCode);
   if (!font) {
-    __glcRaiseError(GLC_RESOURCE_ERROR);
     __glcFree(node);
     return NULL;
   }
   node->data = font;
   FT_List_Add(&inContext->fontList, node);
 
-  /* Update the last font ID (see glcGenFontID())*/
-  if (inFontID >= inContext->lastFontID)
-    inContext->lastFontID = inFontID + 1;
   return font;
 }
 
@@ -1042,10 +1024,15 @@ GLint APIENTRY glcNewFontFromMaster(GLint inFont, GLint inMaster)
   __GLCcontext *ctx = NULL;
   __GLCmaster *master = NULL;
   __GLCfont *font = NULL;
-  FT_ListNode node = NULL;
 
   GLC_INIT_THREAD();
 
+  /* Check if inFont is in legal bounds */
+  if (inFont < 1) {
+    __glcRaiseError(GLC_PARAMETER_ERROR);
+    return 0;
+  }
+
   /* Verify that the thread has a current context and that the master
    * identified by 'inMaster' exists.
    */
@@ -1055,25 +1042,8 @@ GLint APIENTRY glcNewFontFromMaster(GLint inFont, GLint inMaster)
 
   ctx = GLC_GET_CURRENT_CONTEXT();
 
-  /* Check if inFont is in legal bounds */
-  if (inFont < 1) {
-    __glcRaiseError(GLC_PARAMETER_ERROR);
-    return 0;
-  }
-
-  /* Look for the font which ID is inFont in the GLC_FONT_LIST */
-  for (node = ctx->fontList.head; node; node = node->next) {
-    font = (__GLCfont*)node->data;
-    if (font->id == inFont)
-      break;
-  }
-
-  /* No font with an ID equal to inFont has been found */
-  if (!node)
-    font = NULL;
-
   /* Create and return the new font */
-  font = __glcNewFontFromMaster(font, inFont, master, ctx);
+  font = __glcNewFontFromMaster(inFont, master, ctx, 0);
   __glcMasterDestroy(master);
   return font->id;
 }
@@ -1107,7 +1077,6 @@ GLint APIENTRY glcNewFontFromFamily(GLint inFont, const GLCchar* inFamily)
   GLCchar8* UinFamily = NULL;
   __GLCmaster* master = NULL;
   __GLCfont *font = NULL;
-  FT_ListNode node = NULL;
 
   GLC_INIT_THREAD();
 
@@ -1128,10 +1097,8 @@ GLint APIENTRY glcNewFontFromFamily(GLint inFont, const GLCchar* inFamily)
 
   /* Convert the family name in UTF-8 encoding */
   UinFamily = __glcConvertToUtf8(inFamily, ctx->stringState.stringType);
-  if (!UinFamily) {
-    __glcRaiseError(GLC_RESOURCE_ERROR);
+  if (!UinFamily)
     return 0;
-  }
 
   master = __glcMasterFromFamily(ctx, UinFamily);
 
@@ -1139,19 +1106,8 @@ GLint APIENTRY glcNewFontFromFamily(GLint inFont, const GLCchar* inFamily)
   if (!master)
     return 0;
 
-  /* Look for the font which ID is inFont in the GLC_FONT_LIST */
-  for (node = ctx->fontList.head; node; node = node->next) {
-    font = (__GLCfont*)node->data;
-    if (font->id == inFont)
-      break;
-  }
-
-  /* No font with an ID equal to inFont has been found */
-  if (!node)
-    font = NULL;
-
   /* Create and return the new font */
-  font = __glcNewFontFromMaster(font, inFont, master, ctx);
+  font = __glcNewFontFromMaster(inFont, master, ctx, 0);
   __glcMasterDestroy(master);
   return font->id;
 }
