@@ -1,6 +1,6 @@
 /* QuesoGLC
  * A free implementation of the OpenGL Character Renderer (GLC)
- * Copyright (c) 2002, 2004-2007, Bertrand Coconnier
+ * Copyright (c) 2002, 2004-2008, Bertrand Coconnier
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -16,21 +16,20 @@
  *  License along with this library; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
-/* $Id: ocontext.c,v 1.83 2007/10/25 19:53:03 bcoconni Exp $ */
+/* $Id: ocontext.c 741 2008-03-02 13:10:53Z bcoconni $ */
 
 /** \file
  * defines the object __GLCcontext which is used to manage the contexts.
  */
 
-#include "internal.h"
-
-#include <sys/stat.h>
-#ifndef __WIN32__
-#include <unistd.h>
-#else
+#if defined(__WIN32__) || defined(_MSC_VER)
 #include <io.h>
+#else
+#include <unistd.h>
 #endif
+#include <sys/stat.h>
 
+#include "internal.h"
 #include "texture.h"
 #include FT_MODULE_H
 
@@ -67,7 +66,7 @@ __GLCcontext* __glcContextCreate(GLint inContext)
 
   FT_Add_Default_Modules(This->library);
 
-#ifdef FT_CACHE_H
+#ifdef GLC_FT_CACHE
   if (FTC_Manager_New(This->library, 0, 0, 0, __glcFileOpen, NULL,
 		      &This->cache)) {
     __glcRaiseError(GLC_RESOURCE_ERROR);
@@ -80,7 +79,7 @@ __GLCcontext* __glcContextCreate(GLint inContext)
   This->config = FcInitLoadConfigAndFonts();
   if (!This->config) {
     __glcRaiseError(GLC_RESOURCE_ERROR);
-#ifdef FT_CACHE_H
+#ifdef GLC_FT_CACHE
     FTC_Manager_Done(This->cache);
 #endif
     FT_Done_Library(This->library);
@@ -94,8 +93,7 @@ __GLCcontext* __glcContextCreate(GLint inContext)
 
   This->catalogList = __glcArrayCreate(sizeof(GLCchar8*));
   if (!This->catalogList) {
-    __glcRaiseError(GLC_RESOURCE_ERROR);
-#ifdef FT_CACHE_H
+#ifdef GLC_FT_CACHE
     FTC_Manager_Done(This->cache);
 #endif
     FT_Done_Library(This->library);
@@ -107,8 +105,7 @@ __GLCcontext* __glcContextCreate(GLint inContext)
   This->masterHashTable = __glcArrayCreate(sizeof(GLCchar32));
   if (!This->masterHashTable) {
     __glcArrayDestroy(This->catalogList);
-    __glcRaiseError(GLC_RESOURCE_ERROR);
-#ifdef FT_CACHE_H
+#ifdef GLC_FT_CACHE
     FTC_Manager_Done(This->cache);
 #endif
     FT_Done_Library(This->library);
@@ -122,6 +119,8 @@ __GLCcontext* __glcContextCreate(GLint inContext)
   This->currentFontList.tail = NULL;
   This->fontList.head = NULL;
   This->fontList.tail = NULL;
+  This->genFontList.head = NULL;
+  This->genFontList.tail = NULL;
 
   This->isCurrent = GL_FALSE;
   This->isInGlobalCommand = GL_FALSE;
@@ -150,8 +149,7 @@ __GLCcontext* __glcContextCreate(GLint inContext)
   if (!This->measurementBuffer) {
     __glcArrayDestroy(This->masterHashTable);
     __glcArrayDestroy(This->catalogList);
-    __glcRaiseError(GLC_RESOURCE_ERROR);
-#ifdef FT_CACHE_H
+#ifdef GLC_FT_CACHE
     FTC_Manager_Done(This->cache);
 #endif
     FT_Done_Library(This->library);
@@ -162,14 +160,12 @@ __GLCcontext* __glcContextCreate(GLint inContext)
   This->isInCallbackFunc = GL_FALSE;
   This->buffer = NULL;
   This->bufferSize = 0;
-  This->lastFontID = 1;
   This->vertexArray = __glcArrayCreate(2 * sizeof(GLfloat));
   if (!This->vertexArray) {
     __glcArrayDestroy(This->measurementBuffer);
     __glcArrayDestroy(This->masterHashTable);
     __glcArrayDestroy(This->catalogList);
-    __glcRaiseError(GLC_RESOURCE_ERROR);
-#ifdef FT_CACHE_H
+#ifdef GLC_FT_CACHE
     FTC_Manager_Done(This->cache);
 #endif
     FT_Done_Library(This->library);
@@ -178,13 +174,12 @@ __GLCcontext* __glcContextCreate(GLint inContext)
     return NULL;
   }
   This->controlPoints = __glcArrayCreate(5 * sizeof(GLfloat));
-  if (!This->vertexArray) {
+  if (!This->controlPoints) {
     __glcArrayDestroy(This->vertexArray);
     __glcArrayDestroy(This->measurementBuffer);
     __glcArrayDestroy(This->masterHashTable);
     __glcArrayDestroy(This->catalogList);
-    __glcRaiseError(GLC_RESOURCE_ERROR);
-#ifdef FT_CACHE_H
+#ifdef GLC_FT_CACHE
     FTC_Manager_Done(This->cache);
 #endif
     FT_Done_Library(This->library);
@@ -199,8 +194,7 @@ __GLCcontext* __glcContextCreate(GLint inContext)
     __glcArrayDestroy(This->measurementBuffer);
     __glcArrayDestroy(This->masterHashTable);
     __glcArrayDestroy(This->catalogList);
-    __glcRaiseError(GLC_RESOURCE_ERROR);
-#ifdef FT_CACHE_H
+#ifdef GLC_FT_CACHE
     FTC_Manager_Done(This->cache);
 #endif
     FT_Done_Library(This->library);
@@ -217,8 +211,7 @@ __GLCcontext* __glcContextCreate(GLint inContext)
     __glcArrayDestroy(This->masterHashTable);
     __glcArrayDestroy(This->endContour);
     __glcArrayDestroy(This->catalogList);
-    __glcRaiseError(GLC_RESOURCE_ERROR);
-#ifdef FT_CACHE_H
+#ifdef GLC_FT_CACHE
     FTC_Manager_Done(This->cache);
 #endif
     FT_Done_Library(This->library);
@@ -236,8 +229,7 @@ __GLCcontext* __glcContextCreate(GLint inContext)
     __glcArrayDestroy(This->endContour);
     __glcArrayDestroy(This->vertexIndices);
     __glcArrayDestroy(This->catalogList);
-    __glcRaiseError(GLC_RESOURCE_ERROR);
-#ifdef FT_CACHE_H
+#ifdef GLC_FT_CACHE
     FTC_Manager_Done(This->cache);
 #endif
     FT_Done_Library(This->library);
@@ -254,6 +246,7 @@ __GLCcontext* __glcContextCreate(GLint inContext)
   This->atlas.id = 0;
   This->atlas.width = 0;
   This->atlas.heigth = 0;
+  This->atlas.bufferObjectID = 0;
   This->atlasList.head = NULL;
   This->atlasList.tail = NULL;
   This->atlasWidth = 0;
@@ -268,20 +261,9 @@ __GLCcontext* __glcContextCreate(GLint inContext)
     char *path = NULL;
     char *begin = NULL;
     char *sepPos = NULL;
-    char *separator = NULL;
-
-    /* Read the paths of fonts file.
-     * First, try GLC_CATALOG_LIST...
-     */
-    if (getenv("GLC_CATALOG_LIST"))
-      path = strdup(getenv("GLC_CATALOG_LIST"));
-    else if (getenv("GLC_PATH")) {
-      /* Try GLC_PATH which uses the same format than PATH */
-      path = strdup(getenv("GLC_PATH"));
-    }
+    char *separator = getenv("GLC_LIST_SEPARATOR");
 
     /* Get the list separator */
-    separator = getenv("GLC_LIST_SEPARATOR");
     if (!separator) {
 #ifdef __WIN32__
 	/* Windows can not use a colon-separated list since the colon sign is
@@ -297,6 +279,24 @@ __GLCcontext* __glcContextCreate(GLint inContext)
 #endif
     }
 
+    /* Read the paths of fonts file.
+     * First, try GLC_CATALOG_LIST...
+     */
+    if (getenv("GLC_CATALOG_LIST"))
+#ifdef __WIN32__
+      path = _strdup(getenv("GLC_CATALOG_LIST"));
+#else
+      path = strdup(getenv("GLC_CATALOG_LIST"));
+#endif
+    else if (getenv("GLC_PATH")) {
+      /* Try GLC_PATH which uses the same format than PATH */
+#ifdef __WIN32__
+      path = _strdup(getenv("GLC_PATH"));
+#else
+      path = strdup(getenv("GLC_PATH"));
+#endif
+    }
+
     if (path) {
       /* Get each path and add the corresponding masters to the current
        * context */
@@ -309,15 +309,17 @@ __GLCcontext* __glcContextCreate(GLint inContext)
         if (*sepPos)
 	  *(sepPos++) = 0;
 
+#ifdef __WIN32__
+        dup = (GLCchar8*)_strdup(begin);
+#else
         dup = (GLCchar8*)strdup(begin);
+#endif
         if (!dup) {
 	  __glcRaiseError(GLC_RESOURCE_ERROR);
         }
         else {
-	  if (!__glcArrayAppend(This->catalogList, &dup)) {
-            __glcRaiseError(GLC_RESOURCE_ERROR);
+	  if (!__glcArrayAppend(This->catalogList, &dup))
             free(dup);
-          }
           else if (!FcConfigAppFontAddDir(This->config,
                                           (const unsigned char*)begin)) {
             __glcArrayRemove(This->catalogList,
@@ -343,6 +345,22 @@ __GLCcontext* __glcContextCreate(GLint inContext)
 
 
 
+#ifndef GLC_FT_CACHE
+/* This function is called from FT_List_Finalize() to close all the fonts
+ * of the GLC_CURRENT_FONT_LIST
+ */
+static void __glcFontClosure(FT_Memory GLC_UNUSED_ARG(inMemory), void* inData,
+			     void* GLC_UNUSED_ARG(inUser))
+{
+  __GLCfont *font = (__GLCfont*)inData;
+
+  assert(font);
+  __glcFontClose(font);
+}
+#endif
+
+
+
 /* This function is called from FT_List_Finalize() to destroy all
  * remaining fonts
  */
@@ -353,9 +371,8 @@ static void __glcFontDestructor(FT_Memory GLC_UNUSED_ARG(inMemory),
   __GLCcontext* ctx = (__GLCcontext*)inUser;
 
   assert(ctx);
-
-  if (font)
-    __glcFontDestroy(font, ctx);
+  assert(font);
+  __glcFontDestroy(font, ctx);
 }
 
 
@@ -385,12 +402,20 @@ void __glcContextDestroy(__GLCcontext *This)
   __glcArrayDestroy(This->catalogList);
 
   /* Destroy GLC_CURRENT_FONT_LIST */
+#ifdef GLC_FT_CACHE
   FT_List_Finalize(&This->currentFontList, NULL,
 		   &__glcCommonArea.memoryManager, NULL);
+#else
+  FT_List_Finalize(&This->currentFontList, __glcFontClosure,
+		   &__glcCommonArea.memoryManager, NULL);
+#endif
 
   /* Destroy GLC_FONT_LIST */
   FT_List_Finalize(&This->fontList, __glcFontDestructor,
                    &__glcCommonArea.memoryManager, This);
+  /* Destroy empty fonts generated by glcGenFontID() */
+  FT_List_Finalize(&This->genFontList, __glcFontDestructor,
+		   &__glcCommonArea.memoryManager, This);
 
   if (This->masterHashTable)
     __glcArrayDestroy(This->masterHashTable);
@@ -419,7 +444,7 @@ void __glcContextDestroy(__GLCcontext *This)
   if (This->geomBatches)
     __glcArrayDestroy(This->geomBatches);
 
-#ifdef FT_CACHE_H
+#ifdef GLC_FT_CACHE
   FTC_Manager_Done(This->cache);
 #endif
   FT_Done_Library(This->library);
@@ -528,7 +553,7 @@ __GLCfont* __glcContextGetFont(__GLCcontext *This, GLint inCode)
     master = __glcMasterMatchCode(This, inCode);
     if (!master)
       return NULL;
-    font = __glcNewFontFromMaster(NULL, glcGenFontID(), master, This);
+    font = __glcNewFontFromMaster(glcGenFontID(), master, This, inCode);
     __glcMasterDestroy(master);
 
     if (font) {
@@ -591,7 +616,7 @@ static void __glcContextUpdateHashTable(__GLCcontext *This)
     return;
   }
   objectSet = FcObjectSetBuild(FC_FAMILY, FC_FOUNDRY, FC_OUTLINE, FC_SPACING,
-			       FC_CHARSET, NULL);
+			       NULL);
   if (!objectSet) {
     __glcRaiseError(GLC_RESOURCE_ERROR);
     FcPatternDestroy(pattern);
@@ -614,6 +639,9 @@ static void __glcContextUpdateHashTable(__GLCcontext *This)
     int length = GLC_ARRAY_LENGTH(This->masterHashTable);
     GLCchar32* hashTable = (GLCchar32*)GLC_ARRAY_DATA(This->masterHashTable);
     FcBool outline = FcFalse;
+    FcChar8* family = NULL;
+    int fixed = 0;
+    FcChar8* foundry = NULL;
     FcResult result = FcResultMatch;
 
     /* Check whether the glyphs are outlines */
@@ -622,8 +650,30 @@ static void __glcContextUpdateHashTable(__GLCcontext *This)
     if (!outline)
       continue;
 
+    result = FcPatternGetString(fontSet->fonts[i], FC_FAMILY, 0, &family);
+    assert(result != FcResultTypeMismatch);
+    result = FcPatternGetString(fontSet->fonts[i], FC_FOUNDRY, 0, &foundry);
+    assert(result != FcResultTypeMismatch);
+    result = FcPatternGetInteger(fontSet->fonts[i], FC_SPACING, 0, &fixed);
+    assert(result != FcResultTypeMismatch);
+
+    if (foundry)
+      pattern = FcPatternBuild(NULL, FC_FAMILY, FcTypeString, family,
+			       FC_FOUNDRY, FcTypeString, foundry, FC_SPACING,
+			       FcTypeInteger, fixed, NULL);
+    else
+      pattern = FcPatternBuild(NULL, FC_FAMILY, FcTypeString, family,
+			       FC_SPACING, FcTypeInteger, fixed, NULL);
+
+    if (!pattern) {
+      __glcRaiseError(GLC_RESOURCE_ERROR);
+      FcFontSetDestroy(fontSet);
+      return;
+    }
+
     /* Check if the font is already registered in the hash table */
-    hashValue = FcPatternHash(fontSet->fonts[i]);
+    hashValue = FcPatternHash(pattern);
+    FcPatternDestroy(pattern);
     for (j = 0; j < length; j++) {
       if (hashTable[j] == hashValue)
 	break;
@@ -635,7 +685,6 @@ static void __glcContextUpdateHashTable(__GLCcontext *This)
 
     /* Register the font (i.e. append its hash value to the hash table) */
     if (!__glcArrayAppend(This->masterHashTable, &hashValue)) {
-      __glcRaiseError(GLC_RESOURCE_ERROR);
       FcFontSetDestroy(fontSet);
       return;
     }
@@ -649,7 +698,11 @@ static void __glcContextUpdateHashTable(__GLCcontext *This)
 /* Append a catalog to the context catalog list */
 void __glcContextAppendCatalog(__GLCcontext* This, const GLCchar* inCatalog)
 {
+#ifdef __WIN32__
+  GLCchar8* dup = (GLCchar8*)_strdup(inCatalog);
+#else
   GLCchar8* dup = (GLCchar8*)strdup(inCatalog);
+#endif
 
   if (!dup) {
     __glcRaiseError(GLC_RESOURCE_ERROR);
@@ -657,7 +710,6 @@ void __glcContextAppendCatalog(__GLCcontext* This, const GLCchar* inCatalog)
   }
 
   if (!__glcArrayAppend(This->catalogList, &dup)) {
-    __glcRaiseError(GLC_RESOURCE_ERROR);
     free(dup);
     return;
   }
@@ -676,7 +728,11 @@ void __glcContextAppendCatalog(__GLCcontext* This, const GLCchar* inCatalog)
 /* Prepend a catalog to the context catalog list */
 void __glcContextPrependCatalog(__GLCcontext* This, const GLCchar* inCatalog)
 {
+#ifdef __WIN32__
+  GLCchar8* dup = (GLCchar8*)_strdup(inCatalog);
+#else
   GLCchar8* dup = (GLCchar8*)strdup(inCatalog);
+#endif
 
   if (!dup) {
     __glcRaiseError(GLC_RESOURCE_ERROR);
@@ -684,7 +740,6 @@ void __glcContextPrependCatalog(__GLCcontext* This, const GLCchar* inCatalog)
   }
 
   if (!__glcArrayInsert(This->catalogList, 0, &dup)) {
-    __glcRaiseError(GLC_RESOURCE_ERROR);
     free(dup);
     return;
   }
@@ -751,18 +806,17 @@ void __glcContextRemoveCatalog(__GLCcontext* This, GLint inIndex)
    */
   for (node = This->fontList.head; node; node = node->next) {
     __GLCfont* font = (__GLCfont*)(node->data);
-    __GLCmaster* master = __glcMasterCreate(font->parentMasterID, This);
     GLCchar32 hashValue = 0;
     GLCchar32* hashTable = (GLCchar32*)GLC_ARRAY_DATA(This->masterHashTable);
     int length = GLC_ARRAY_LENGTH(This->masterHashTable);
     int i = 0;
+    __GLCmaster* master = __glcMasterCreate(font->parentMasterID, This);
+
+    if (!master)
+      continue;
 
-    if (!master) {
-      __glcRaiseError(GLC_RESOURCE_ERROR);
-      return;
-    }
     /* Check if the hash value of the master is in the hash table */
-    hashValue = FcPatternHash(master->pattern);
+    hashValue = GLC_MASTER_HASH_VALUE(master);
     for (i = 0; i < length; i++) {
       if (hashValue == hashTable[i])
 	break;
