@@ -1,6 +1,6 @@
 /* QuesoGLC
  * A free implementation of the OpenGL Character Renderer (GLC)
- * Copyright (c) 2002, 2004-2007, Bertrand Coconnier
+ * Copyright (c) 2002, 2004-2008, Bertrand Coconnier
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -16,7 +16,7 @@
  *  License along with this library; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
-/* $Id: ofacedesc.c,v 1.25 2007/10/21 16:49:14 bcoconni Exp $ */
+/* $Id: ofacedesc.c 710 2008-01-21 20:14:17Z bcoconni $ */
 
 /** \file
  * defines the object __GLCfaceDescriptor that contains the description of a
@@ -31,7 +31,7 @@
 
 #include "internal.h"
 #include FT_GLYPH_H
-#ifdef FT_CACHE_H
+#ifdef GLC_FT_CACHE
 #include FT_CACHE_H
 #endif
 #include FT_OUTLINE_H
@@ -56,53 +56,88 @@
  */
 __GLCfaceDescriptor* __glcFaceDescCreate(__GLCmaster* inMaster,
 					 const GLCchar8* inFace,
-					 __GLCcontext* inContext)
+					 __GLCcontext* inContext, GLint inCode)
 {
   __GLCfaceDescriptor* This = NULL;
   FcObjectSet* objectSet = NULL;
   FcFontSet *fontSet = NULL;
-  FcPattern* pattern = inMaster->pattern;
   int i = 0;
+  FcPattern* pattern = FcPatternCreate();
 
-  if (inFace) {
-    pattern = FcPatternDuplicate(inMaster->pattern);
-    if (!pattern) {
-      __glcRaiseError(GLC_RESOURCE_ERROR);
-      return NULL;
-    }
-
-    if (!FcPatternAddString(pattern, FC_STYLE, inFace)) {
-      __glcRaiseError(GLC_RESOURCE_ERROR);
-      FcPatternDestroy(pattern);
-      return NULL;
-    }
+  if (!pattern) {
+    __glcRaiseError(GLC_RESOURCE_ERROR);
+    return NULL;
   }
 
-  objectSet = FcObjectSetBuild(FC_STYLE, FC_SPACING, FC_FILE, FC_INDEX,
-			       FC_OUTLINE, NULL);
+  objectSet = FcObjectSetBuild(FC_FAMILY, FC_FOUNDRY, FC_STYLE, FC_SPACING,
+			       FC_FILE, FC_INDEX, FC_OUTLINE, FC_CHARSET, NULL);
   if (!objectSet) {
     __glcRaiseError(GLC_RESOURCE_ERROR);
+    FcPatternDestroy(pattern);
     return NULL;
   }
   fontSet = FcFontList(inContext->config, pattern, objectSet);
   FcObjectSetDestroy(objectSet);
+  FcPatternDestroy(pattern);
   if (!fontSet) {
     __glcRaiseError(GLC_RESOURCE_ERROR);
+    FcPatternDestroy(pattern);
     return NULL;
   }
 
-  if (inFace)
-    FcPatternDestroy(pattern);
-
   for (i = 0; i < fontSet->nfont; i++) {
+    FcChar8* family = NULL;
+    int fixed = 0;
+    FcChar8* foundry = NULL;
+    FcChar8* style = NULL;
     FcBool outline = FcFalse;
+    FcCharSet* charSet = NULL;
     FcResult result = FcResultMatch;
+    FcBool equal = FcFalse;
+
+    result = FcPatternGetCharSet(fontSet->fonts[i], FC_CHARSET, 0, &charSet);
+    assert(result != FcResultTypeMismatch);
+    if (inCode && !FcCharSetHasChar(charSet, inCode))
+      continue;
 
     /* Check whether the glyphs are outlines */
     result = FcPatternGetBool(fontSet->fonts[i], FC_OUTLINE, 0, &outline);
     assert(result != FcResultTypeMismatch);
-    if (outline)
+    if (!outline)
+      continue;
+
+    result = FcPatternGetString(fontSet->fonts[i], FC_FAMILY, 0, &family);
+    assert(result != FcResultTypeMismatch);
+    result = FcPatternGetString(fontSet->fonts[i], FC_FOUNDRY, 0, &foundry);
+    assert(result != FcResultTypeMismatch);
+    result = FcPatternGetInteger(fontSet->fonts[i], FC_SPACING, 0, &fixed);
+    assert(result != FcResultTypeMismatch);
+
+    if (foundry)
+      pattern = FcPatternBuild(NULL, FC_FAMILY, FcTypeString, family,
+			       FC_FOUNDRY, FcTypeString, foundry, FC_SPACING,
+			       FcTypeInteger, fixed, NULL);
+    else
+      pattern = FcPatternBuild(NULL, FC_FAMILY, FcTypeString, family,
+			       FC_SPACING, FcTypeInteger, fixed, NULL);
+
+    if (!pattern) {
+      __glcRaiseError(GLC_RESOURCE_ERROR);
+      FcFontSetDestroy(fontSet);
+      return NULL;
+    }
+
+    equal = FcPatternEqual(pattern, inMaster->pattern);
+    FcPatternDestroy(pattern);
+    if (equal) {
+      if (inFace) {
+	result = FcPatternGetString(fontSet->fonts[i], FC_STYLE, 0, &style);
+	assert(result != FcResultTypeMismatch);
+	if (strcmp((const char*)style, (const char*)inFace))
+	  continue;
+      }
       break;
+    }
   }
 
   if (i == fontSet->nfont) {
@@ -129,8 +164,8 @@ __GLCfaceDescriptor* __glcFaceDescCreate(__GLCmaster* inMaster,
   This->node.prev = NULL;
   This->node.next = NULL;
   This->node.data = NULL;
-#ifndef FT_CACHE_H
   This->face = NULL;
+#ifndef GLC_FT_CACHE
   This->faceRefCount = 0;
 #endif
   This->glyphList.head = NULL;
@@ -147,6 +182,10 @@ void __glcFaceDescDestroy(__GLCfaceDescriptor* This, __GLCcontext* inContext)
   FT_ListNode node = NULL;
   FT_ListNode next = NULL;
 
+#ifndef GLC_FT_CACHE
+  assert(!This->faceRefCount && !This->face);
+#endif
+
   /* Don't use FT_List_Finalize here, since __glcGlyphDestroy also destroys
    * the node itself.
    */
@@ -157,7 +196,7 @@ void __glcFaceDescDestroy(__GLCfaceDescriptor* This, __GLCcontext* inContext)
     node = next;
   }
 
-#if defined(FT_CACHE_H) \
+#if defined(GLC_FT_CACHE) \
   && (FREETYPE_MAJOR > 2 \
      || (FREETYPE_MAJOR == 2 \
          && (FREETYPE_MINOR > 1 \
@@ -172,7 +211,7 @@ void __glcFaceDescDestroy(__GLCfaceDescriptor* This, __GLCcontext* inContext)
 
 
 
-#ifndef FT_CACHE_H
+#ifndef GLC_FT_CACHE
 /* Open a face, select a Unicode charmap. __glcFaceDesc maintains a reference
  * count for each face so that the face is open only once.
  */
@@ -199,13 +238,7 @@ FT_Face __glcFaceDescOpen(__GLCfaceDescriptor* This,
     }
 
     /* select a Unicode charmap */
-    if (FT_Select_Charmap(This->face, ft_encoding_unicode)) {
-      /* No Unicode charmap is available */
-      __glcRaiseError(GLC_RESOURCE_ERROR);
-      FT_Done_Face(This->face);
-      This->face = NULL;
-      return NULL;
-    }
+    FT_Select_Charmap(This->face, ft_encoding_unicode);
 
     This->faceRefCount = 1;
   }
@@ -234,7 +267,7 @@ void __glcFaceDescClose(__GLCfaceDescriptor* This)
 
 
 
-#else /* FT_CACHE_H */
+#else /* GLC_FT_CACHE */
 /* Callback function used by the FreeType cache manager to open a given face */
 FT_Error __glcFileOpen(FTC_FaceID inFile, FT_Library inLibrary,
 		       FT_Pointer GLC_UNUSED_ARG(inData), FT_Face* outFace)
@@ -260,13 +293,11 @@ FT_Error __glcFileOpen(FTC_FaceID inFile, FT_Library inLibrary,
   }
 
   /* select a Unicode charmap */
-  error = FT_Select_Charmap(*outFace, ft_encoding_unicode);
-  if (error)
-    __glcRaiseError(GLC_RESOURCE_ERROR);
+  FT_Select_Charmap(*outFace, ft_encoding_unicode);
 
   return error;
 }
-#endif /* FT_CACHE_H */
+#endif /* GLC_FT_CACHE */
 
 
 
@@ -286,7 +317,7 @@ __GLCglyph* __glcFaceDescGetGlyph(__GLCfaceDescriptor* This, GLint inCode,
   }
 
   /* Open the face */
-#ifdef FT_CACHE_H
+#ifdef GLC_FT_CACHE
   if (FTC_Manager_LookupFace(inContext->cache, (FTC_FaceID)This, &face)) {
 #else
   face = __glcFaceDescOpen(This, inContext);
@@ -297,7 +328,7 @@ __GLCglyph* __glcFaceDescGetGlyph(__GLCfaceDescriptor* This, GLint inCode,
   }
 
   /* Create a new glyph */
-#ifdef FT_CACHE_H
+#ifdef GLC_FT_CACHE
   glyph = __glcGlyphCreate(FT_Get_Char_Index(face, inCode), inCode);
   if (!glyph)
     return NULL;
@@ -312,7 +343,7 @@ __GLCglyph* __glcFaceDescGetGlyph(__GLCfaceDescriptor* This, GLint inCode,
    * face.
    */
   FT_List_Add(&This->glyphList, (FT_ListNode)glyph);
-#ifndef FT_CACHE_H
+#ifndef GLC_FT_CACHE
   __glcFaceDescClose(This);
 #endif
   return glyph;
@@ -320,18 +351,16 @@ __GLCglyph* __glcFaceDescGetGlyph(__GLCfaceDescriptor* This, GLint inCode,
 
 
 
-/* Load a FreeType glyph (FT_Glyph) of the current face and returns the 
- * corresponding FT_Face. The size of the glyph is given by inScaleX and
+/* Load a glyph of the current font face and stores the corresponding data in
+ * the corresponding face. The size of the glyph is given by inScaleX and
  * inScaleY. 'inGlyphIndex' contains the index of the glyph in the font file.
  */
-static FT_Face __glcFaceDescLoadFreeTypeGlyph(__GLCfaceDescriptor* This,
-				       __GLCcontext* inContext,
-				       GLfloat inScaleX, GLfloat inScaleY,
-				       GLCulong inGlyphIndex)
+GLboolean __glcFaceDescPrepareGlyph(__GLCfaceDescriptor* This,
+				    __GLCcontext* inContext, GLfloat inScaleX,
+				    GLfloat inScaleY, GLCulong inGlyphIndex)
 {
-  FT_Face face = NULL;
   FT_Int32 loadFlags = FT_LOAD_NO_BITMAP | FT_LOAD_IGNORE_TRANSFORM;
-#ifdef FT_CACHE_H
+#ifdef GLC_FT_CACHE
 # if FREETYPE_MAJOR == 2 \
      && (FREETYPE_MINOR < 1 \
          || (FREETYPE_MINOR == 1 && FREETYPE_PATCH < 8))
@@ -349,7 +378,7 @@ static FT_Face __glcFaceDescLoadFreeTypeGlyph(__GLCfaceDescriptor* This,
     loadFlags |= FT_LOAD_NO_HINTING;
 
   /* Open the face */
-#ifdef FT_CACHE_H
+#ifdef GLC_FT_CACHE
 # if FREETYPE_MAJOR == 2 \
      && (FREETYPE_MINOR < 1 \
          || (FREETYPE_MINOR == 1 && FREETYPE_PATCH < 8))
@@ -361,53 +390,52 @@ static FT_Face __glcFaceDescLoadFreeTypeGlyph(__GLCfaceDescriptor* This,
       (inContext->renderState.resolution < GLC_EPSILON ?
        72. : inContext->renderState.resolution) / 72.);
 
-  if (FTC_Manager_Lookup_Size(inContext->cache, &font, &face, &size)) {
+  if (FTC_Manager_Lookup_Size(inContext->cache, &font, &This->face, &size)) {
     __glcRaiseError(GLC_RESOURCE_ERROR);
-    return NULL;
+    return GL_FALSE;
   }
 # else
   scaler.face_id = (FTC_FaceID)This;
-  scaler.width = (FT_F26Dot6)(inScaleX * 64.);
-  scaler.height = (FT_F26Dot6)(inScaleY * 64.);
-  scaler.pixel = 0;
-  scaler.x_res = (GLCuint)inContext->renderState.resolution;
-  scaler.y_res = (GLCuint)inContext->renderState.resolution;
+  scaler.width = (FT_UInt)(inScaleX * 64.);
+  scaler.height = (FT_UInt)(inScaleY * 64.);
+  scaler.pixel = (FT_Int)0;
+  scaler.x_res = (FT_UInt)(inContext->renderState.resolution < GLC_EPSILON ?
+			   72 : inContext->renderState.resolution);
+  scaler.y_res = (FT_UInt)(inContext->renderState.resolution < GLC_EPSILON ?
+			   72 : inContext->renderState.resolution);
 
   if (FTC_Manager_LookupSize(inContext->cache, &scaler, &size)) {
     __glcRaiseError(GLC_RESOURCE_ERROR);
-    return NULL;
+    return GL_FALSE;
   }
 
-  face = size->face;
+  This->face = size->face;
 # endif /* FREETYPE_MAJOR */
 #else
-  face = __glcFaceDescOpen(This, inContext);
-  if (!face) {
-    __glcRaiseError(GLC_RESOURCE_ERROR);
-    return NULL;
-  }
+  if (!__glcFaceDescOpen(This, inContext))
+    return GL_FALSE;
 
   /* Select the size of the glyph */
-  if (FT_Set_Char_Size(face, (FT_F26Dot6)(inScaleX * 64.),
+  if (FT_Set_Char_Size(This->face, (FT_F26Dot6)(inScaleX * 64.),
 		       (FT_F26Dot6)(inScaleY * 64.),
-		       (GLCuint)inContext->resolution,
-		       (GLCuint)inContext->resolution)) {
+		       (FT_UInt)inContext->renderState.resolution,
+		       (FT_UInt)inContext->renderState.resolution)) {
     __glcFaceDescClose(This);
     __glcRaiseError(GLC_RESOURCE_ERROR);
-    return NULL;
+    return GL_FALSE;
   }
 #endif
 
   /* Load the glyph */
-  if (FT_Load_Glyph(face, inGlyphIndex, loadFlags)) {
+  if (FT_Load_Glyph(This->face, inGlyphIndex, loadFlags)) {
     __glcRaiseError(GLC_RESOURCE_ERROR);
-#ifndef FT_CACHE_H
+#ifndef GLC_FT_CACHE
     __glcFaceDescClose(This);
 #endif
-    return NULL;
+    return GL_FALSE;
   }
 
-  return face;
+  return GL_TRUE;
 }
 
 
@@ -438,16 +466,15 @@ GLfloat* __glcFaceDescGetBoundingBox(__GLCfaceDescriptor* This,
 {
   FT_BBox boundBox;
   FT_Glyph glyph;
-  FT_Face face = __glcFaceDescLoadFreeTypeGlyph(This, inContext, inScaleX,
-						inScaleY, inGlyphIndex);
 
   assert(outVec);
 
-  if (!face)
+  if (!__glcFaceDescPrepareGlyph(This, inContext, inScaleX, inScaleY,
+				 inGlyphIndex))
     return NULL;
 
   /* Get the bounding box of the glyph */
-  FT_Get_Glyph(face->glyph, &glyph);
+  FT_Get_Glyph(This->face->glyph, &glyph);
   FT_Glyph_Get_CBox(glyph, ft_glyph_bbox_unscaled, &boundBox);
 
   /* Transform the bounding box according to the conversion from FT_F26Dot6 to
@@ -459,7 +486,7 @@ GLfloat* __glcFaceDescGetBoundingBox(__GLCfaceDescriptor* This,
   outVec[3] = (GLfloat) boundBox.yMax / 64. / inScaleY;
 
   FT_Done_Glyph(glyph);
-#ifndef FT_CACHE_H
+#ifndef GLC_FT_CACHE
   __glcFaceDescClose(This);
 #endif
   return outVec;
@@ -471,36 +498,25 @@ GLfloat* __glcFaceDescGetBoundingBox(__GLCfaceDescriptor* This,
  * inScaleY. The result is returned in outVec. 'inGlyphIndex' contains the
  * index of the glyph in the font file.
  */
-#ifdef FT_CACHE_H
 GLfloat* __glcFaceDescGetAdvance(__GLCfaceDescriptor* This,
 				 GLCulong inGlyphIndex, GLfloat* outVec,
 				 GLfloat inScaleX, GLfloat inScaleY,
 				 __GLCcontext* inContext)
-#else
-GLfloat* __glcFaceDescGetAdvance(__GLCfaceDescriptor* This,
-				 GLCulong inGlyphIndex, GLfloat* outVec,
-				 GLfloat inScaleX, GLfloat inScaleY,
-                                 GLboolean inCloseFile,
-				 __GLCcontext* inContext)
-#endif
 {
-  FT_Face face = __glcFaceDescLoadFreeTypeGlyph(This, inContext, inScaleX,
-						inScaleY, inGlyphIndex);
-
   assert(outVec);
 
-  if (!face)
+  if (!__glcFaceDescPrepareGlyph(This, inContext, inScaleX, inScaleY,
+				 inGlyphIndex))
     return NULL;
 
   /* Transform the advance according to the conversion from FT_F26Dot6 to
    * GLfloat.
    */
-  outVec[0] = (GLfloat) face->glyph->advance.x / 64. / inScaleX;
-  outVec[1] = (GLfloat) face->glyph->advance.y / 64. / inScaleY;
+  outVec[0] = (GLfloat) This->face->glyph->advance.x / 64. / inScaleX;
+  outVec[1] = (GLfloat) This->face->glyph->advance.y / 64. / inScaleY;
 
-#ifndef FT_CACHE_H
-  if (inCloseFile)
-    __glcFaceDescClose(This);
+#ifndef GLC_FT_CACHE
+  __glcFaceDescClose(This);
 #endif
   return outVec;
 }
@@ -535,7 +551,7 @@ GLCchar8* __glcFaceDescGetFontFormat(__GLCfaceDescriptor* This,
   GLCchar8* result = NULL;
 
   /* Open the face */
-#ifdef FT_CACHE_H
+#ifdef GLC_FT_CACHE
   if (FTC_Manager_LookupFace(inContext->cache, (FTC_FaceID)This, &face)) {
 #else
   face = __glcFaceDescOpen(This, inContext);
@@ -552,9 +568,9 @@ GLCchar8* __glcFaceDescGetFontFormat(__GLCfaceDescriptor* This,
      * the existence of FT_XFREE86_H is checked.
      */
     result = (GLCchar8*)FT_Get_X11_Font_Format(face);
-#  ifndef FT_CACHE_H
+#  ifndef GLC_FT_CACHE
     __glcFaceDescClose(This);
-#  endif /* FT_CACHE_H */
+#  endif /* GLC_FT_CACHE */
     return result;
   }
 #endif /* FT_XFREE86_H */
@@ -564,6 +580,9 @@ GLCchar8* __glcFaceDescGetFontFormat(__GLCfaceDescriptor* This,
     switch(inAttrib) {
 #ifndef FT_XFREE86_H
     case GLC_MASTER_FORMAT:
+#ifndef GLC_FT_CACHE
+      __glcFaceDescClose(This);
+#endif
       return masterFormat1;
 #endif
     case GLC_FULL_NAME_SGI:
@@ -653,7 +672,7 @@ GLCchar8* __glcFaceDescGetFontFormat(__GLCfaceDescriptor* This,
 #endif
   }
 
-#ifndef FT_CACHE_H
+#ifndef GLC_FT_CACHE
   /* Close the face */
   __glcFaceDescClose(This);
 #endif
@@ -675,7 +694,7 @@ GLfloat* __glcFaceDescGetMaxMetric(__GLCfaceDescriptor* This, GLfloat* outVec,
 
   assert(outVec);
 
-#ifdef FT_CACHE_H
+#ifdef GLC_FT_CACHE
   if (FTC_Manager_LookupFace(inContext->cache, (FTC_FaceID)This, &face))
 #else
   face = __glcFaceDescOpen(This, inContext);
@@ -693,7 +712,7 @@ GLfloat* __glcFaceDescGetMaxMetric(__GLCfaceDescriptor* This, GLfloat* outVec,
   outVec[4] = (GLfloat)face->bbox.xMax * scale;
   outVec[5] = (GLfloat)face->bbox.xMin * scale;
 
-#ifndef FT_CACHE_H
+#ifndef GLC_FT_CACHE
   __glcFaceDescClose(This);
 #endif
   return outVec;
@@ -711,24 +730,23 @@ GLfloat* __glcFaceDescGetKerning(__GLCfaceDescriptor* This,
 {
   FT_Vector kerning;
   FT_Error error;
-  FT_Face face = __glcFaceDescLoadFreeTypeGlyph(This, inContext, inScaleX,
-						inScaleY, inGlyphIndex);
 
   assert(outVec);
 
-  if (!face)
+  if (!__glcFaceDescPrepareGlyph(This, inContext, inScaleX, inScaleY,
+				 inGlyphIndex))
     return NULL;
 
-  if (!FT_HAS_KERNING(face)) {
+  if (!FT_HAS_KERNING(This->face)) {
     outVec[0] = 0.;
     outVec[1] = 0.;
     return outVec;
   }
 
-  error = FT_Get_Kerning(face, inPrevGlyphIndex, inGlyphIndex,
+  error = FT_Get_Kerning(This->face, inPrevGlyphIndex, inGlyphIndex,
 			 FT_KERNING_DEFAULT, &kerning);
 
-#ifndef FT_CACHE_H
+#ifndef GLC_FT_CACHE
   __glcFaceDescClose(This);
 #endif
 
@@ -785,10 +803,8 @@ static int __glcMoveTo(FT_Vector *inVecTo, void* inUserData)
    */
 
   if (!__glcArrayAppend(data->endContour,
-			&GLC_ARRAY_LENGTH(data->vertexArray))) {
-    __glcRaiseError(GLC_RESOURCE_ERROR);
+			&GLC_ARRAY_LENGTH(data->vertexArray)))
     return 1;
-  }
 
   data->vector[0] = (GLfloat) inVecTo->x;
   data->vector[1] = (GLfloat) inVecTo->y;
@@ -809,10 +825,8 @@ static int __glcLineTo(FT_Vector *inVecTo, void* inUserData)
 {
   __GLCrendererData *data = (__GLCrendererData *) inUserData;
 
-  if (!__glcArrayAppend(data->vertexArray, data->vector)) {
-    __glcRaiseError(GLC_RESOURCE_ERROR);
+  if (!__glcArrayAppend(data->vertexArray, data->vector))
     return 1;
-  }
 
   data->vector[0] = (GLfloat) inVecTo->x;
   data->vector[1] = (GLfloat) inVecTo->y;
@@ -893,7 +907,7 @@ GLboolean __glcFaceDescOutlineDecompose(__GLCfaceDescriptor* This,
   FT_Outline_Funcs outlineInterface;
   FT_Face face = NULL;
 
-#ifndef FT_CACHE_H
+#ifndef GLC_FT_CACHE
   face = This->face;
 #else
   if (FTC_Manager_LookupFace(inContext->cache, (FTC_FaceID)This, &face)) {
@@ -953,21 +967,12 @@ GLboolean __glcFaceDescGetBitmapSize(__GLCfaceDescriptor* This, GLint* outWidth,
                                      GLfloat inScaleX, GLfloat inScaleY,
                                      int inFactor, __GLCcontext* inContext)
 {
-  FT_Face face = NULL;
   FT_Outline outline;
   FT_Matrix matrix;
   FT_BBox boundingBox;
+  FT_Face face = This->face;
 
-
-#ifndef FT_CACHE_H
-  face = This->face;
   assert(face);
-#else
-  if (FTC_Manager_LookupFace(inContext->cache, (FTC_FaceID)This, &face)) {
-    __glcRaiseError(GLC_RESOURCE_ERROR);
-    return GL_FALSE;
-  }
-#endif
 
   outline = face->glyph->outline;
 
@@ -991,7 +996,7 @@ GLboolean __glcFaceDescGetBitmapSize(__GLCfaceDescriptor* This, GLint* outWidth,
     boundingBox.xMax = (boundingBox.xMax + 63) & -64;	/* ceiling(xMax) */
     boundingBox.yMax = (boundingBox.yMax + 63) & -64;	/* ceiling(yMax) */
 
-    /* Calculate pitch to upper 8 byte boundary for 1 bit / pixel, i.e. ceil() */
+    /* Calculate pitch to upper 8 byte boundary for 1 bit/pixel, i.e. ceil() */
     pitch = (boundingBox.xMax - boundingBox.xMin + 511) >> 9;
     *outWidth = pitch << 3;
     *outHeight = (boundingBox.yMax - boundingBox.yMin) >> 6;
@@ -1045,21 +1050,13 @@ GLboolean __glcFaceDescGetBitmap(__GLCfaceDescriptor* This, GLint inWidth,
                                  GLint inHeight, void* inBuffer,
                                  __GLCcontext* inContext)
 {
-  FT_Face face = NULL;
   FT_Outline outline;
   FT_BBox boundingBox;
   FT_Bitmap pixmap;
   FT_Matrix matrix;
+  FT_Face face = This->face;
 
-#ifndef FT_CACHE_H
-  face = This->face;
   assert(face);
-#else
-  if (FTC_Manager_LookupFace(inContext->cache, (FTC_FaceID)This, &face)) {
-    __glcRaiseError(GLC_RESOURCE_ERROR);
-    return GL_FALSE;
-  }
-#endif
 
   outline = face->glyph->outline;
   FT_Outline_Get_CBox(&outline, &boundingBox);
@@ -1069,7 +1066,7 @@ GLboolean __glcFaceDescGetBitmap(__GLCfaceDescriptor* This, GLint inWidth,
   pixmap.buffer = inBuffer;
 
   if (inContext->renderState.renderStyle == GLC_BITMAP) {
-    /* Calculate pitch to upper 8 byte boundary for 1 bit / pixel, i.e. ceil() */
+    /* Calculate pitch to upper 8 byte boundary for 1 bit/pixel, i.e. ceil() */
     pixmap.pitch = -(pixmap.width >> 3);
 
     /* Fill the pixmap descriptor and the pixmap buffer */
@@ -1077,7 +1074,7 @@ GLboolean __glcFaceDescGetBitmap(__GLCfaceDescriptor* This, GLint inWidth,
   }
   else {
     /* Flip the picture */
-    pixmap.pitch = -pixmap.width; /* 8 bits / pixel */
+    pixmap.pitch = -pixmap.width; /* 8 bits/pixel */
 
     /* Fill the pixmap descriptor and the pixmap buffer */
     pixmap.pixel_mode = ft_pixel_mode_grays; /* Anti-aliased rendering */
@@ -1123,20 +1120,36 @@ GLboolean __glcFaceDescGetBitmap(__GLCfaceDescriptor* This, GLint inWidth,
 GLboolean __glcFaceDescOutlineEmpty(__GLCfaceDescriptor* This,
                                     __GLCcontext* inContext)
 {
-  FT_Face face = NULL;
-  FT_Outline outline;
-
-#ifndef FT_CACHE_H
-  face = This->face;
-  assert(face);
-#else
-  if (FTC_Manager_LookupFace(inContext->cache, (FTC_FaceID)This, &face)) {
-    __glcRaiseError(GLC_RESOURCE_ERROR);
-    return GL_FALSE;
-  }
-#endif
-
-  outline = face->glyph->outline;
+  FT_Outline outline = This->face->glyph->outline;
 
   return outline.n_points ? GL_TRUE : GL_FALSE;
 }
+
+
+
+/* Get the CharMap of the face descriptor */
+__GLCcharMap* __glcFaceDescGetCharMap(__GLCfaceDescriptor* This,
+				      __GLCcontext* inContext)
+{
+  FcResult result = FcResultMatch;
+  FcCharSet* charSet = NULL;
+  FcCharSet* newCharSet = NULL;
+  __GLCcharMap* charMap = __glcCharMapCreate(NULL, inContext);
+
+  if (!charMap)
+    return NULL;
+
+  result = FcPatternGetCharSet(This->pattern, FC_CHARSET, 0, &charSet);
+  assert(result != FcResultTypeMismatch);
+
+  newCharSet = FcCharSetCopy(charSet);
+  if (!newCharSet) {
+    __glcCharMapDestroy(charMap);
+    return NULL;
+  }
+
+  FcCharSetDestroy(charMap->charSet);
+  charMap->charSet = newCharSet;
+
+  return charMap;
+}
