@@ -1,6 +1,6 @@
 /* QuesoGLC
  * A free implementation of the OpenGL Character Renderer (GLC)
- * Copyright (c) 2002, 2004-2007, Bertrand Coconnier
+ * Copyright (c) 2002, 2004-2008, Bertrand Coconnier
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -16,7 +16,7 @@
  *  License along with this library; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
-/* $Id: render.c,v 1.95 2007/10/28 13:38:20 bcoconni Exp $ */
+/* $Id: render.c 749 2008-03-08 20:18:29Z bcoconni $ */
 
 /** \file
  * defines the so-called "Rendering commands" described in chapter 3.9 of the
@@ -107,8 +107,8 @@ static void __glcRenderCharBitmap(__GLCfont* inFont, __GLCcontext* inContext,
   void* pixBuffer = NULL;
   GLint boundingBox[4] = {0, 0, 0, 0};
 
-  __glcFaceDescGetBitmapSize(inFont->faceDesc, &pixWidth, &pixHeight,
-                             boundingBox, scale_x, scale_y, 0, inContext);
+  __glcFontGetBitmapSize(inFont, &pixWidth, &pixHeight, boundingBox, scale_x,
+			 scale_y, 0, inContext);
 
   pixBuffer = (GLubyte *)__glcMalloc(pixWidth * pixHeight);
   if (!pixBuffer) {
@@ -117,10 +117,8 @@ static void __glcRenderCharBitmap(__GLCfont* inFont, __GLCcontext* inContext,
   }
 
   /* render the glyph */
-  if (!__glcFaceDescGetBitmap(inFont->faceDesc, pixWidth, pixHeight,
-                              pixBuffer, inContext)) {
+  if (!__glcFontGetBitmap(inFont, pixWidth, pixHeight, pixBuffer, inContext)) {
     __glcFree(pixBuffer);
-    __glcRaiseError(GLC_RESOURCE_ERROR);
     return;
   }
 
@@ -176,6 +174,11 @@ static void* __glcRenderChar(GLint inCode, GLint inPrevCode, GLboolean inIsRTL,
   if ((scale_x == 0.f) || (scale_y == 0.f))
     return NULL;
 
+#ifndef GLC_FT_CACHE
+  if (!__glcFontOpen(inFont, inContext))
+    return NULL;
+#endif
+
   if (inPrevCode && inContext->enableState.kerning) {
     GLfloat kerning[2];
     GLint leftCode = inIsRTL ? inCode : inPrevCode;
@@ -194,77 +197,49 @@ static void* __glcRenderChar(GLint inCode, GLint inPrevCode, GLboolean inIsRTL,
 		 + kerning[1] * inContext->bitmapMatrix[3],
 		 NULL);
       else
-	glTranslatef(kerning[0], kerning[1], 0.);
+	glTranslatef(kerning[0], kerning[1], 0.f);
     }
   }
 
+  if (!__glcFontGetAdvance(inFont, inCode, advance, inContext, scale_x,
+			   scale_y)) {
+#ifndef GLC_FT_CACHE
+    __glcFontClose(inFont);
+#endif
+    return NULL;
+  }
+
   /* Get and load the glyph which unicode code is identified by inCode */
   glyph = __glcFontGetGlyph(inFont, inCode, inContext);
 
-  /* Renders the display lists if they exist and if GLC_GL_OBJECTS is enabled
-   * then return.
-   */
-  if (inContext->enableState.glObjects) {
-    if (inIsRTL)
-      glTranslatef(-glyph->advance[0], glyph->advance[1], 0.);
-
-    switch(inContext->renderState.renderStyle) {
-    case GLC_TEXTURE:
-      if (glyph->displayList[0]) {
-	glCallList(glyph->displayList[0]);
-	/* The glyph is put at the head of the atlas list, so that we keep track
-	 * of glyphes that are used often in order not to remove those ones from
-	 * the atlas when it is full.
-	 */
-	FT_List_Up(&inContext->atlasList, (FT_ListNode)glyph->textureObject);
-	if (!inIsRTL)
-	  glTranslatef(glyph->advance[0], glyph->advance[1], 0.);
-	return NULL;
-      }
-      break;
-    case GLC_LINE:
-      if (glyph->displayList[1]) {
-	glCallList(glyph->displayList[1]);
-	if (!inIsRTL)
-	  glTranslatef(glyph->advance[0], glyph->advance[1], 0.);
-	return NULL;
-      }
-      break;
-    case GLC_TRIANGLE:
-      if ((!inContext->enableState.extrude) && glyph->displayList[2]) {
-	glCallList(glyph->displayList[2]);
-	if (!inIsRTL)
-	  glTranslatef(glyph->advance[0], glyph->advance[1], 0.);
-	return NULL;
-      }
-      if (inContext->enableState.extrude && glyph->displayList[3]) {
-	glCallList(glyph->displayList[3]);
-	if (!inIsRTL)
-	  glTranslatef(glyph->advance[0], glyph->advance[1], 0.);
-	return NULL;
-      }
-      break;
-    }
-  }
-
-  if (!__glcFaceDescGetAdvance(inFont->faceDesc, glyph->index, advance, scale_x,
-                               scale_y, inContext))
+  if (inContext->enableState.glObjects
+      && !__glcFontPrepareGlyph(inFont, inContext, scale_x, scale_y,
+			     glyph->index)) {
+#ifndef GLC_FT_CACHE
+    __glcFontClose(inFont);
+#endif
     return NULL;
+  }
 
   sx64 = 64. * scale_x;
   sy64 = 64. * scale_y;
 
   if (inContext->renderState.renderStyle != GLC_BITMAP) {
     if (inIsRTL)
-      glTranslatef(-advance[0], advance[1], 0.);
+      glTranslatef(-advance[0], advance[1], 0.f);
 
     /* If the outline contains no point then the glyph represents a space
      * character and there is no need to continue the process of rendering.
      */
-    if (!__glcFaceDescOutlineEmpty(inFont->faceDesc, inContext)) {
+    if (!__glcFontOutlineEmpty(inFont, inContext)) {
       /* Update the advance and return */
       if (!inIsRTL)
-        glTranslatef(advance[0], advance[1], 0.);
+        glTranslatef(advance[0], advance[1], 0.f);
+      if (inContext->enableState.glObjects)
+	glyph->isSpacingChar = GL_TRUE;
+#ifndef GLC_FT_CACHE
+      __glcFontClose(inFont);
+#endif
       return NULL;
     }
 
@@ -272,11 +247,7 @@ static void* __glcRenderChar(GLint inCode, GLint inPrevCode, GLboolean inIsRTL,
      * divide the scale by 2^6
      */
     if (!inContext->enableState.glObjects)
-      glScalef(1. / sx64, 1. / sy64, 1.);
-    else {
-      glyph->advance[0] = advance[0];
-      glyph->advance[1] = advance[1];
-    }
+      glScalef(1. / sx64, 1. / sy64, 1.f);
   }
 
   /* Call the appropriate function depending on the rendering mode. It first
@@ -285,18 +256,19 @@ static void* __glcRenderChar(GLint inCode, GLint inPrevCode, GLboolean inIsRTL,
    */
   switch(inContext->renderState.renderStyle) {
   case GLC_BITMAP:
-    __glcRenderCharBitmap(inFont, inContext, scale_x, scale_y, advance, inIsRTL);
+    __glcRenderCharBitmap(inFont, inContext, scale_x, scale_y, advance,
+			  inIsRTL);
     break;
   case GLC_TEXTURE:
-      __glcRenderCharTexture(inFont, inContext, scale_x, scale_y, glyph);
+    __glcRenderCharTexture(inFont, inContext, scale_x, scale_y, glyph);
     break;
   case GLC_LINE:
-      __glcRenderCharScalable(inFont, inContext, GLC_LINE, transformMatrix,
-			      scale_x, scale_y, glyph);
+    __glcRenderCharScalable(inFont, inContext, transformMatrix, scale_x,
+			    scale_y, glyph);
     break;
   case GLC_TRIANGLE:
-      __glcRenderCharScalable(inFont, inContext, GLC_TRIANGLE, transformMatrix,
-			      scale_x, scale_y, glyph);
+    __glcRenderCharScalable(inFont, inContext, transformMatrix, scale_x,
+			    scale_y, glyph);
     break;
   default:
     __glcRaiseError(GLC_PARAMETER_ERROR);
@@ -308,14 +280,234 @@ static void* __glcRenderChar(GLint inCode, GLint inPrevCode, GLboolean inIsRTL,
     if (!inIsRTL)
       glTranslatef(advance[0], advance[1], 0.);
   }
-#ifndef FT_CACHE_H
-  __glcFaceDescClose(inFont->faceDesc);
+#ifndef GLC_FT_CACHE
+  __glcFontClose(inFont);
 #endif
   return NULL;
 }
 
 
 
+/* This internal function is used by both glcRenderString() and
+ * glcRenderCountedString(). The string 'inString' must be sorted in visual
+ * order and stored using UCS4 format.
+ */
+static void __glcRenderCountedString(__GLCcontext* inContext, GLCchar* inString,
+				     GLboolean inIsRightToLeft,GLint inCount)
+{
+  GLint listIndex = 0;
+  GLint i = 0;
+  GLCchar32* ptr = NULL;
+  __GLCglState GLState;
+  __GLCcharacter prevCode = {0, NULL, NULL, {0.f, 0.f}};
+  GLboolean saveGLObjects = GL_FALSE;
+  GLint shift = 1;
+  __GLCcharacter* chars = NULL;
+
+  /* Disable the internal management of GL objects when the user is currently
+   * building a display list.
+   */
+  glGetIntegerv(GL_LIST_INDEX, &listIndex);
+  if (listIndex) {
+    saveGLObjects = inContext->enableState.glObjects;
+    inContext->enableState.glObjects = GL_FALSE;
+  }
+
+  if (inContext->enableState.glObjects
+      && inContext->renderState.renderStyle != GLC_BITMAP) {
+    chars = (__GLCcharacter*)__glcMalloc(inCount * sizeof(__GLCcharacter));
+    if (!chars) {
+      __glcRaiseError(GLC_RESOURCE_ERROR);
+      return;
+    }
+  }
+
+  /* Save the value of the GL parameters */
+  __glcSaveGLState(&GLState, inContext, GL_FALSE);
+
+  if (inContext->renderState.renderStyle == GLC_LINE ||
+      inContext->renderState.renderStyle == GLC_TRIANGLE) {
+    glPushClientAttrib(GL_CLIENT_VERTEX_ARRAY_BIT);
+    glEnableClientState(GL_VERTEX_ARRAY);
+    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
+    glDisableClientState(GL_COLOR_ARRAY);
+    glDisableClientState(GL_INDEX_ARRAY);
+    glDisableClientState(GL_NORMAL_ARRAY);
+    glDisableClientState(GL_EDGE_FLAG_ARRAY);
+  }
+
+  /* Set the texture environment if the render style is GLC_TEXTURE */
+  if (inContext->renderState.renderStyle == GLC_TEXTURE) {
+    /* Set the new values of the parameters */
+    glEnable(GL_BLEND);
+    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
+    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
+    if (inContext->enableState.glObjects) {
+      if (inContext->atlas.id)
+	glBindTexture(GL_TEXTURE_2D, inContext->atlas.id);
+      if (GLEW_ARB_vertex_buffer_object) {
+	glPushClientAttrib(GL_CLIENT_VERTEX_ARRAY_BIT);
+	if (inContext->atlas.bufferObjectID) {
+	  glBindBufferARB(GL_ARRAY_BUFFER_ARB, inContext->atlas.bufferObjectID);
+	  glInterleavedArrays(GL_T2F_V3F, 0, NULL);
+	}
+      }
+    }
+    else if (inContext->texture.id) {
+      glBindTexture(GL_TEXTURE_2D, inContext->texture.id);
+      if (GLEW_ARB_pixel_buffer_object && inContext->texture.bufferObjectID)
+	glBindBufferARB(GL_PIXEL_UNPACK_BUFFER,
+			inContext->texture.bufferObjectID);
+    }
+  }
+
+  /* Render the string */
+  ptr = inString;
+  if (inIsRightToLeft) {
+    ptr += inCount - 1;
+    shift = -1;
+  }
+
+  if (inContext->enableState.glObjects
+      && inContext->renderState.renderStyle != GLC_BITMAP) {
+    __GLCfont* font = NULL;
+    __GLCglyph* glyph = NULL;
+    int length = 0;
+    int j = 0;
+    GLuint GLObjectIndex = inContext->renderState.renderStyle - 0x101;
+    FT_ListNode node = NULL;
+
+    if (inContext->renderState.renderStyle == GLC_TRIANGLE
+	&& inContext->enableState.extrude)
+      GLObjectIndex++;
+
+    for (i = 0; i < inCount; i++) {
+      if (*ptr >= 32) {
+ 	for (node = inContext->currentFontList.head; node ; node = node->next) {
+ 	  font = (__GLCfont*)node->data;
+ 	  glyph = __glcCharMapGetGlyph(font->charMap, *ptr);
+
+ 	  if (glyph) {
+ 	    if (!glyph->glObject[GLObjectIndex] && !glyph->isSpacingChar)
+ 	      continue;
+
+	    if (!glyph->isSpacingChar
+		&& (inContext->renderState.renderStyle == GLC_TEXTURE))
+	      FT_List_Up(&inContext->atlasList,
+			 (FT_ListNode)glyph->textureObject);
+
+	    chars[length].glyph = glyph;
+	    chars[length].advance[0] = glyph->advance[0];
+	    chars[length].advance[1] = glyph->advance[1];
+
+ 	    if (inContext->enableState.kerning) {
+ 	      if (prevCode.code && prevCode.font == font) {
+ 		GLfloat kerning[2];
+ 		GLint leftCode = inIsRightToLeft ? *ptr : prevCode.code;
+ 		GLint rightCode = inIsRightToLeft ? prevCode.code : *ptr;
+
+ 		if (__glcFontGetKerning(font, leftCode, rightCode, kerning,
+ 					inContext, GLC_POINT_SIZE,
+ 					GLC_POINT_SIZE)) {
+ 		  if (inIsRightToLeft)
+ 		    chars[length].advance[0] -= kerning[0];
+ 		  else
+ 		    chars[length].advance[0] += kerning[0];
+
+ 		  chars[length].advance[1] += kerning[1];
+ 		}
+ 	      }
+ 	    }
+
+	    prevCode.font = font;
+	    prevCode.code = *ptr;
+
+	    if (glyph->isSpacingChar)
+	      chars[length].code = 32;
+	    else
+	      chars[length].code = *ptr;
+
+ 	    length++;
+ 	    break;
+ 	  }
+ 	}
+      }
+
+      if(!node || (i == inCount-1)) {
+	for (j = 0; j < length; j++) {
+	  if (inIsRightToLeft)
+	    glTranslatef(-chars[j].advance[0], chars[j].advance[1], 0.);
+	  if (chars[j].code != 32) {
+	    __GLCglyph* glyph = chars[j].glyph;
+
+	    switch(inContext->renderState.renderStyle) {
+	    case GLC_TEXTURE:
+	      if (GLEW_ARB_vertex_buffer_object) {
+		glNormal3f(0.f, 0.f, 1.f);
+		glDrawArrays(GL_QUADS, glyph->textureObject->position * 4, 4);
+	      }
+	      else
+		glCallList(glyph->glObject[1]);
+	      break;
+	    case GLC_LINE:
+	      if (GLEW_ARB_vertex_buffer_object && glyph->glObject[0]) {
+		int k = 0;
+
+		glBindBufferARB(GL_ARRAY_BUFFER_ARB, glyph->glObject[0]);
+		glVertexPointer(2, GL_FLOAT, 0, NULL);
+		glNormal3f(0.f, 0.f, 1.f);
+		for (k = 0; k < glyph->nContour; k++)
+		  glDrawArrays(GL_LINE_LOOP, glyph->contours[k],
+			       glyph->contours[k+1] - glyph->contours[k]);
+		break;
+	      }
+	      glCallList(glyph->glObject[0]);
+	      break;
+	    case GLC_TRIANGLE:
+	      glCallList(glyph->glObject[GLObjectIndex]);
+	      break;
+	    }
+	  }
+	  if (!inIsRightToLeft)
+	    glTranslatef(chars[j].advance[0], chars[j].advance[1], 0.);
+	}
+
+	if (!node)
+	  __glcProcessChar(inContext, *ptr, &prevCode, inIsRightToLeft,
+			   __glcRenderChar, NULL);
+	length = 0;
+      }
+
+      ptr += shift;
+    }
+  }
+  else {
+    for (i = 0; i < inCount; i++) {
+      if (*ptr >= 32)
+	__glcProcessChar(inContext, *ptr, &prevCode, inIsRightToLeft,
+			 __glcRenderChar, NULL);
+      ptr += shift;
+    }
+  }
+
+  /* Restore the values of the GL state if needed */
+  __glcRestoreGLState(&GLState, inContext, GL_FALSE);
+
+  if (inContext->renderState.renderStyle != GLC_BITMAP) {
+    if (inContext->enableState.glObjects)
+      __glcFree(chars);
+    if (inContext->renderState.renderStyle != GLC_TEXTURE)
+      glPopClientAttrib();
+    else if (inContext->enableState.glObjects && GLEW_ARB_vertex_buffer_object)
+      glPopClientAttrib();
+  }
+
+  if (listIndex)
+    inContext->enableState.glObjects = saveGLObjects;
+}
+
+
+
 /** \ingroup render
  *  This command renders the character that \e inCode is mapped to.
  *  \param inCode The character to render
@@ -329,10 +521,6 @@ void APIENTRY glcRenderChar(GLint inCode)
 {
   __GLCcontext *ctx = NULL;
   GLint code = 0;
-  __GLCglState GLState;
-  __GLCcharacter prevCode = { 0, NULL };
-  GLint listIndex = 0;
-  GLboolean saveGLObjects = GL_FALSE;
 
   GLC_INIT_THREAD();
 
@@ -348,51 +536,7 @@ void APIENTRY glcRenderChar(GLint inCode)
   if (code < 32)
     return; /* Skip control characters and unknown characters */
 
-  /* Disable the internal management of GL objects when the user is currently
-   * building a display list.
-   */
-  glGetIntegerv(GL_LIST_INDEX, &listIndex);
-  if (listIndex) {
-    saveGLObjects = ctx->enableState.glObjects;
-    ctx->enableState.glObjects = GL_FALSE;
-  }
-
-  /* Save the value of the GL parameters */
-  __glcSaveGLState(&GLState, ctx, GL_FALSE);
-
-  if (ctx->renderState.renderStyle == GLC_LINE ||
-      ctx->renderState.renderStyle == GLC_TRIANGLE) {
-    glEnableClientState(GL_VERTEX_ARRAY);
-    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
-    glDisableClientState(GL_COLOR_ARRAY);
-    glDisableClientState(GL_INDEX_ARRAY);
-    glDisableClientState(GL_NORMAL_ARRAY);
-    glDisableClientState(GL_EDGE_FLAG_ARRAY);
-  }
-
-  /* Set the texture environment if the render style is GLC_TEXTURE */
-  if (ctx->renderState.renderStyle == GLC_TEXTURE) {
-    /* Set the new values of the parameters */
-    glEnable(GL_TEXTURE_2D);
-    glEnable(GL_BLEND);
-    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
-    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
-    if (ctx->enableState.glObjects && ctx->atlas.id)
-      glBindTexture(GL_TEXTURE_2D, ctx->atlas.id);
-    else if (!ctx->enableState.glObjects && ctx->texture.id) {
-      glBindTexture(GL_TEXTURE_2D, ctx->texture.id);
-      if (GLEW_ARB_pixel_buffer_object && ctx->texture.bufferObjectID)
-	glBindBufferARB(GL_PIXEL_UNPACK_BUFFER,
-			ctx->texture.bufferObjectID);
-    }
-  }
-
-  __glcProcessChar(ctx, code, &prevCode, GL_FALSE, __glcRenderChar, NULL);
-
-  /* Restore the values of the GL state if needed */
-  __glcRestoreGLState(&GLState, ctx, GL_FALSE);
-  if (listIndex)
-    ctx->enableState.glObjects = saveGLObjects;
+  __glcRenderCountedString(ctx, (GLCchar*)&code, GL_FALSE, 1);
 }
 
 
@@ -412,15 +556,9 @@ void APIENTRY glcRenderChar(GLint inCode)
  */
 void APIENTRY glcRenderCountedString(GLint inCount, const GLCchar *inString)
 {
-  GLint i = 0;
   __GLCcontext *ctx = NULL;
   GLCchar32* UinString = NULL;
-  GLCchar32* ptr = NULL;
-  __GLCglState GLState;
-  __GLCcharacter prevCode = { 0, NULL };
   GLboolean isRightToLeft = GL_FALSE;
-  GLint listIndex = 0;
-  GLboolean saveGLObjects = GL_FALSE;
 
   GLC_INIT_THREAD();
 
@@ -446,67 +584,11 @@ void APIENTRY glcRenderCountedString(GLint inCount, const GLCchar *inString)
    */
   UinString = __glcConvertCountedStringToVisualUcs4(ctx, &isRightToLeft,
 						    inString, inCount);
-  if (!UinString) {
-    __glcRaiseError(GLC_RESOURCE_ERROR);
+  if (!UinString)
     return;
-  }
 
-  /* Disable the internal management of GL objects when the user is currently
-   * building a display list.
-   */
-  glGetIntegerv(GL_LIST_INDEX, &listIndex);
-  if (listIndex) {
-    saveGLObjects = ctx->enableState.glObjects;
-    ctx->enableState.glObjects = GL_FALSE;
-  }
 
-  /* Save the value of the GL parameters */
-  __glcSaveGLState(&GLState, ctx, GL_FALSE);
-
-  if (ctx->renderState.renderStyle == GLC_LINE ||
-      ctx->renderState.renderStyle == GLC_TRIANGLE) {
-    glEnableClientState(GL_VERTEX_ARRAY);
-    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
-    glDisableClientState(GL_COLOR_ARRAY);
-    glDisableClientState(GL_INDEX_ARRAY);
-    glDisableClientState(GL_NORMAL_ARRAY);
-    glDisableClientState(GL_EDGE_FLAG_ARRAY);
-  }
-
-  /* Set the texture environment if the render style is GLC_TEXTURE */
-  if (ctx->renderState.renderStyle == GLC_TEXTURE) {
-    /* Set the new values of the parameters */
-    glEnable(GL_TEXTURE_2D);
-    glEnable(GL_BLEND);
-    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
-    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
-    if (ctx->enableState.glObjects && ctx->atlas.id)
-      glBindTexture(GL_TEXTURE_2D, ctx->atlas.id);
-    else if (!ctx->enableState.glObjects && ctx->texture.id) {
-      glBindTexture(GL_TEXTURE_2D, ctx->texture.id);
-      if (GLEW_ARB_pixel_buffer_object && ctx->texture.bufferObjectID)
-	glBindBufferARB(GL_PIXEL_UNPACK_BUFFER,
-			ctx->texture.bufferObjectID);
-    }
-  }
-
-  /* Render the string */
-  ptr = UinString;
-  for (i = 0; i < inCount; i++) {
-    if (*ptr < 32) {
-      /* Skip control characters */
-      ptr++;
-      continue;
-    }
-    else
-      __glcProcessChar(ctx, *(ptr++), &prevCode, isRightToLeft, __glcRenderChar,
-                       NULL);
-  }
-
-  /* Restore the values of the GL state if needed */
-  __glcRestoreGLState(&GLState, ctx, GL_FALSE);
-  if (listIndex)
-    ctx->enableState.glObjects = saveGLObjects;
+  __glcRenderCountedString(ctx, UinString, isRightToLeft, inCount);
 }
 
 
@@ -522,12 +604,8 @@ void APIENTRY glcRenderString(const GLCchar *inString)
 {
   __GLCcontext *ctx = NULL;
   GLCchar32* UinString = NULL;
-  GLCchar32* ptr = NULL;
-  __GLCglState GLState;
-  __GLCcharacter prevCode = { 0, NULL };
   GLboolean isRightToLeft = GL_FALSE;
-  GLint listIndex = 0;
-  GLboolean saveGLObjects = GL_FALSE;
+  GLint length = 0;
 
   GLC_INIT_THREAD();
 
@@ -545,68 +623,11 @@ void APIENTRY glcRenderString(const GLCchar *inString)
   /* Creates a Unicode string based on the current string type. Basically,
    * that means that inString is read in the current string format.
    */
-  UinString = __glcConvertToVisualUcs4(ctx, &isRightToLeft, inString);
-  if (!UinString) {
-    __glcRaiseError(GLC_RESOURCE_ERROR);
+  UinString = __glcConvertToVisualUcs4(ctx, &isRightToLeft, &length, inString);
+  if (!UinString)
     return;
-  }
 
-  /* Disable the internal management of GL objects when the user is currently
-   * building a display list.
-   */
-  glGetIntegerv(GL_LIST_INDEX, &listIndex);
-  if (listIndex) {
-    saveGLObjects = ctx->enableState.glObjects;
-    ctx->enableState.glObjects = GL_FALSE;
-  }
-
-  /* Save the value of the GL parameters */
-  __glcSaveGLState(&GLState, ctx, GL_FALSE);
-
-  if (ctx->renderState.renderStyle == GLC_LINE ||
-      ctx->renderState.renderStyle == GLC_TRIANGLE) {
-    glEnableClientState(GL_VERTEX_ARRAY);
-    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
-    glDisableClientState(GL_COLOR_ARRAY);
-    glDisableClientState(GL_INDEX_ARRAY);
-    glDisableClientState(GL_NORMAL_ARRAY);
-    glDisableClientState(GL_EDGE_FLAG_ARRAY);
-  }
-
-  /* Set the texture environment if the render style is GLC_TEXTURE */
-  if (ctx->renderState.renderStyle == GLC_TEXTURE) {
-    /* Set the new values of the parameters */
-    glEnable(GL_TEXTURE_2D);
-    glEnable(GL_BLEND);
-    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
-    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
-    if (ctx->enableState.glObjects && ctx->atlas.id)
-      glBindTexture(GL_TEXTURE_2D, ctx->atlas.id);
-    else if (!ctx->enableState.glObjects && ctx->texture.id) {
-      glBindTexture(GL_TEXTURE_2D, ctx->texture.id);
-      if (GLEW_ARB_pixel_buffer_object && ctx->texture.bufferObjectID)
-	glBindBufferARB(GL_PIXEL_UNPACK_BUFFER,
-			ctx->texture.bufferObjectID);
-    }
-  }
-
-  /* Render the string */
-  ptr = UinString;
-  while (*ptr) {
-    if (*ptr < 32) {
-      /* Skip control characters */
-      ptr++;
-      continue;
-    }
-    else
-      __glcProcessChar(ctx, *(ptr++), &prevCode, isRightToLeft, __glcRenderChar,
-                       NULL);
-  }
-
-  /* Restore the values of the GL state if needed */
-  __glcRestoreGLState(&GLState, ctx, GL_FALSE);
-  if (listIndex)
-    ctx->enableState.glObjects = saveGLObjects;
+  __glcRenderCountedString(ctx, UinString, isRightToLeft, length);
 }
 
 
