@@ -1,6 +1,6 @@
 /* QuesoGLC
  * A free implementation of the OpenGL Character Renderer (GLC)
- * Copyright (c) 2002, 2004-2007, Bertrand Coconnier
+ * Copyright (c) 2002, 2004-2008, Bertrand Coconnier
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -16,7 +16,7 @@
  *  License along with this library; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
-/* $Id: texture.c,v 1.19 2007/10/27 12:02:19 bcoconni Exp $ */
+/* $Id: texture.c 748 2008-03-08 20:15:34Z bcoconni $ */
 
 /** \file
  *  defines the routines used to render characters with textures.
@@ -164,6 +164,23 @@ static GLboolean __glcTextureAtlasGetPosition(__GLCcontext* inContext,
   atlasNode->glyph = inGlyph;
   inGlyph->textureObject = atlasNode;
 
+  if (GLEW_ARB_vertex_buffer_object) {
+    /* Create a VBO, if none exists yet */
+    if (!inContext->atlas.bufferObjectID) {
+      glGenBuffersARB(1, &inContext->atlas.bufferObjectID);
+      if (!inContext->atlas.bufferObjectID) {
+	__glcRaiseError(GLC_RESOURCE_ERROR);
+	/* Even though we failed to create a VBO ID, the rendering of the glyph
+	 * can be processed without VBO, so we return GL_TRUE.
+	 */
+	return GL_TRUE;
+      }
+    }
+    /* Bind the buffer and define/update its size */
+    glBindBufferARB(GL_ARRAY_BUFFER_ARB, inContext->atlas.bufferObjectID);
+  }
+
+
   return GL_TRUE;
 }
 
@@ -185,6 +202,10 @@ static GLboolean __glcTextureGetImmediate(__GLCcontext* inContext,
 	|| (inHeight > inContext->texture.heigth)) {
       /* The texture is not large enough so we destroy the current texture */
       glDeleteTextures(1, &inContext->texture.id);
+      inWidth = (inWidth > inContext->texture.width) ?
+	inWidth : inContext->texture.width;
+      inHeight = (inHeight > inContext->texture.heigth) ?
+	inHeight : inContext->texture.heigth;
       inContext->texture.id = 0;
       inContext->texture.width = 0;
       inContext->texture.heigth = 0;
@@ -277,7 +298,7 @@ void __glcRenderCharTexture(__GLCfont* inFont, __GLCcontext* inContext,
       return;
 
     /* Compute the size of the pixmap where the glyph will be rendered */
-    atlasNode = (__GLCatlasElement*)inGlyph->textureObject;
+    atlasNode = inGlyph->textureObject;
 
     __glcFaceDescGetBitmapSize(inFont->faceDesc, &pixWidth, &pixHeight,
                                boundingBox, scale_x, scale_y, 0, inContext);
@@ -327,7 +348,7 @@ void __glcRenderCharTexture(__GLCfont* inFont, __GLCcontext* inContext,
   do {
     if (GLEW_ARB_pixel_buffer_object && !inContext->enableState.glObjects) {
       pixBuffer = (GLubyte *)glMapBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB,
-						GL_WRITE_ONLY);
+						GL_WRITE_ONLY_ARB);
       if (!pixBuffer) {
 	__glcRaiseError(GLC_RESOURCE_ERROR);
 	return;
@@ -338,11 +359,12 @@ void __glcRenderCharTexture(__GLCfont* inFont, __GLCcontext* inContext,
     if (!__glcFaceDescGetBitmap(inFont->faceDesc, pixWidth, pixHeight,
                                 pixBuffer, inContext)) {
       glPopClientAttrib();
+
       if (GLEW_ARB_pixel_buffer_object && !inContext->enableState.glObjects)
         glUnmapBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB);
       else
         __glcFree(pixBuffer);
-      __glcRaiseError(GLC_RESOURCE_ERROR);
+
       return;
     }
 
@@ -401,32 +423,110 @@ void __glcRenderCharTexture(__GLCfont* inFont, __GLCcontext* inContext,
   width = (GLfloat)((boundingBox[2] - boundingBox[0]) / 64.);
   heigth = (GLfloat)((boundingBox[3] - boundingBox[1]) / 64.);
 
+  if (pixBuffer)
+    __glcFree(pixBuffer);
+
   /* Add the new texture to the texture list and the new display list
    * to the list of display lists
    */
   if (inContext->enableState.glObjects) {
-    inGlyph->displayList[0] = glGenLists(1);
-    if (!inGlyph->displayList[0]) {
-      __glcRaiseError(GLC_RESOURCE_ERROR);
-      if (pixBuffer)
-        __glcFree(pixBuffer);
+    if (GLEW_ARB_vertex_buffer_object) {
+      GLfloat* buffer = NULL;
+      GLfloat* data = NULL;
+      __GLCatlasElement* atlasNode = inGlyph->textureObject;
+
+      buffer = __glcMalloc(inContext->atlasWidth * inContext->atlasHeight * 20
+			   * sizeof(GLfloat));
+      if (!buffer) {
+	__glcRaiseError(GLC_RESOURCE_ERROR);
+	return;
+      }
+
+      /* The display list ID is used as a flag to declare that the VBO has been
+       * initialized and can be used.
+       */
+      inGlyph->glObject[1] = 0xffffffff;
+
+      /* Here we do not use the GL command glBufferSubData() since it seems to
+       * be buggy on some GL drivers (the DRI Intel specifically).
+       * Instead, we use a workaround: the current values of the VBO are stored
+       * in memory and new values are appended to them. Then, the content of
+       * the resulting array replaces all the values previously stored in the
+       * VBO.
+       */
+      if (inContext->atlasCount > 1) {
+	data = glMapBufferARB(GL_ARRAY_BUFFER_ARB, GL_READ_ONLY);
+	if (!data) {
+	  __glcRaiseError(GLC_RESOURCE_ERROR);
+	  __glcFree(buffer);
+	  return;
+	}
+	memcpy(buffer, data, inContext->atlasCount * 20 * sizeof(GLfloat));
+	glUnmapBufferARB(GL_ARRAY_BUFFER_ARB);
+      }
+
+      data = buffer + atlasNode->position * 20;
+
+      data[0] = posX / texWidth;
+      data[1] = posY / texHeigth;
+      data[2] = boundingBox[0] / 64. / GLC_TEXTURE_SIZE;
+      data[3] = boundingBox[1] / 64. / GLC_TEXTURE_SIZE;
+      data[4] = 0.f;
+      data[5] = (posX + width) / texWidth;
+      data[6] = data[1];
+      data[7] = boundingBox[2] / 64. / GLC_TEXTURE_SIZE;
+      data[8] = data[3];
+      data[9] = 0.f;
+      data[10] = data[5];
+      data[11] = (posY + heigth) / texHeigth;
+      data[12] = data[7];
+      data[13] = boundingBox[3] / 64. / GLC_TEXTURE_SIZE;
+      data[14] = 0.f;
+      data[15] = data[0];
+      data[16] = data[11];
+      data[17] = data[2];
+      data[18] = data[13];
+      data[19] = 0.f;
+
+      /* Size of the buffer data is equal to the number of glyphes than can be
+       * stored in the texture times 20 GLfloat (4 vertices made of 3D
+       * coordinates plus 2D texture coordinates : 4 * (3 + 2) = 20)
+       */
+      glBufferDataARB(GL_ARRAY_BUFFER_ARB,
+		      inContext->atlasWidth * inContext->atlasHeight
+		      * 20 * sizeof(GLfloat), buffer, GL_STATIC_DRAW_ARB);
+
+      __glcFree(buffer);
+
+      /* Do the actual GL rendering */
+      glInterleavedArrays(GL_T2F_V3F, 0, NULL);
+      glNormal3f(0.f, 0.f, 1.f);
+      glDrawArrays(GL_QUADS, atlasNode->position * 4, 4);
+
       return;
     }
+    else {
+      inGlyph->glObject[1] = glGenLists(1);
+      if (!inGlyph->glObject[1]) {
+	__glcRaiseError(GLC_RESOURCE_ERROR);
+	return;
+      }
 
-    /* Create the display list */
-    glNewList(inGlyph->displayList[0], GL_COMPILE);
-    glScalef(1. / 64. / scale_x, 1. / 64. / scale_y , 1.);
+      /* Create the display list */
+      glNewList(inGlyph->glObject[1], GL_COMPILE);
+      glScalef(1. / 64. / scale_x, 1. / 64. / scale_y , 1.);
 
-    /* Modify the bouding box dimensions to compensate the glScalef() */
-    boundingBox[0] *= scale_x / GLC_TEXTURE_SIZE;
-    boundingBox[1] *= scale_y / GLC_TEXTURE_SIZE;
-    boundingBox[2] *= scale_x / GLC_TEXTURE_SIZE;
-    boundingBox[3] *= scale_y / GLC_TEXTURE_SIZE;
+      /* Modify the bouding box dimensions to compensate the glScalef() */
+      boundingBox[0] *= scale_x / GLC_TEXTURE_SIZE;
+      boundingBox[1] *= scale_y / GLC_TEXTURE_SIZE;
+      boundingBox[2] *= scale_x / GLC_TEXTURE_SIZE;
+      boundingBox[3] *= scale_y / GLC_TEXTURE_SIZE;
+    }
   }
 
   /* Do the actual GL rendering */
   glBegin(GL_QUADS);
-  glNormal3f(0., 0., 1.);
+  glNormal3f(0.f, 0.f, 1.f);
   glTexCoord2f(posX / texWidth, posY / texHeigth);
   glVertex2i(boundingBox[0], boundingBox[1]);
   glTexCoord2f((posX + width) / texWidth, posY / texHeigth);
@@ -441,9 +541,6 @@ void __glcRenderCharTexture(__GLCfont* inFont, __GLCcontext* inContext,
     /* Finish display list creation */
     glScalef(64. * scale_x, 64. * scale_y, 1.);
     glEndList();
-    glCallList(inGlyph->displayList[0]);
+    glCallList(inGlyph->glObject[1]);
   }
-
-  if (pixBuffer)
-    __glcFree(pixBuffer);
 }
