@@ -1,6 +1,6 @@
 /* QuesoGLC
  * A free implementation of the OpenGL Character Renderer (GLC)
- * Copyright (c) 2002, 2004-2007, Bertrand Coconnier
+ * Copyright (c) 2002, 2004-2008, Bertrand Coconnier
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -16,7 +16,7 @@
  *  License along with this library; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
-/* $Id: unicode.c,v 1.14 2007/10/26 21:06:04 bcoconni Exp $ */
+/* $Id: unicode.c 710 2008-01-21 20:14:17Z bcoconni $ */
 
 /* This file defines miscellaneous utility routines used for Unicode management
  */
@@ -36,12 +36,16 @@ GLCchar* __glcNameFromCode(GLint code)
 {
   GLint position = -1;
 
-  if ((code < 0) || (code > __glcMaxCode))
+  if ((code < 0) || (code > __glcMaxCode)) {
+    __glcRaiseError(GLC_PARAMETER_ERROR);
     return GLC_NONE;
+  }
 
   position = __glcNameFromCodeArray[code];
-  if (position == -1)
+  if (position == -1) {
+    __glcRaiseError(GLC_PARAMETER_ERROR);
     return GLC_NONE;
+  }
 
   return __glcCodeFromNameArray[position].name;
 }
@@ -70,6 +74,8 @@ GLint __glcCodeFromName(GLCchar* name)
     return __glcCodeFromNameArray[start].code;
   if (strcmp(name, __glcCodeFromNameArray[end].name) == 0)
     return __glcCodeFromNameArray[end].code;
+
+  __glcRaiseError(GLC_PARAMETER_ERROR);
   return -1;
 }
 
@@ -283,7 +289,11 @@ GLCchar8* __glcConvertToUtf8(const GLCchar* inString, const GLint inStringType)
     /* If the string is already encoded in UTF-8 format then all we need to do
      * is to make a copy of it.
      */
+#ifdef __WIN32__
+    string = (GLCchar8*)_strdup((const char*)inString);
+#else
     string = (GLCchar8*)strdup((const char*)inString);
+#endif
     break;
   default:
     return NULL;
@@ -322,7 +332,6 @@ GLCchar* __glcConvertFromUtf8ToBuffer(__GLCcontext* This,
 				&len_buffer);
 	if (shift < 0) {
 	  /* There is an ill-formed character in the UTF-8 string, abort */
-	  __glcRaiseError(GLC_PARAMETER_ERROR);
 	  return NULL;
 	}
 	utf8 += shift;
@@ -359,7 +368,6 @@ GLCchar* __glcConvertFromUtf8ToBuffer(__GLCcontext* This,
 				&len_buffer);
 	if (shift < 0) {
 	  /* There is an ill-formed character in the UTF-8 string, abort */
-	  __glcRaiseError(GLC_PARAMETER_ERROR);
 	  return NULL;
 	}
 	utf8 += shift;
@@ -497,7 +505,8 @@ GLint __glcConvertGLintToUcs4(__GLCcontext *inContext, GLint inCode)
     /* Convert the codepoint in UCS4 format and check if it is ill-formed or
      * not
      */
-    if (FcUtf8ToUcs4((GLCchar8*)&inCode, (GLCchar32*)&code, sizeof(GLint)) < 0) {
+    if (FcUtf8ToUcs4((GLCchar8*)&inCode, (GLCchar32*)&code,
+		     sizeof(GLint)) < 0) {
       __glcRaiseError(GLC_PARAMETER_ERROR);
       return -1;
     }
@@ -511,8 +520,9 @@ GLint __glcConvertGLintToUcs4(__GLCcontext *inContext, GLint inCode)
 /* Convert 'inString' (stored in logical order) to UCS4 format and return a
  * copy of the converted string in visual order.
  */
-GLCchar32* __glcConvertToVisualUcs4(__GLCcontext* inContext, GLboolean *outIsRTL,
-				   const GLCchar* inString)
+GLCchar32* __glcConvertToVisualUcs4(__GLCcontext* inContext,
+				    GLboolean *outIsRTL, GLint *outLength,
+				    const GLCchar* inString)
 {
   GLCchar32* string = NULL;
   int length = 0;
@@ -591,7 +601,6 @@ GLCchar32* __glcConvertToVisualUcs4(__GLCcontext* inContext, GLboolean *outIsRTL
 	shift = FcUtf8ToUcs4(utf8, &buffer, strlen((const char*)utf8));
 	if (shift < 0) {
 	  /* There is an ill-formed character in the UTF-8 string, abort */
-	  __glcRaiseError(GLC_PARAMETER_ERROR);
 	  return NULL;
 	}
 	utf8 += shift;
@@ -628,6 +637,8 @@ GLCchar32* __glcConvertToVisualUcs4(__GLCcontext* inContext, GLboolean *outIsRTL
   else
     visualString = string;
 
+  *outLength = length;
+
   return visualString;
 }
 
